import gzip
import importlib.util
import zlib
from typing import (
    AsyncIterator,
    Iterator,
)

import pytest

from zapros import Headers, Response
from zapros._models import (
    AsyncClosableStream,
    ClosableStream,
)


class StreamWrapper(ClosableStream):
    def __init__(self, iterator: Iterator[bytes]) -> None:
        self._iterator = iterator

    def __iter__(
        self,
    ) -> "StreamWrapper":
        return self

    def __next__(self) -> bytes:
        return next(self._iterator)

    def close(self) -> None:
        pass


class AsyncStreamWrapper(AsyncClosableStream):
    def __init__(
        self,
        generator: AsyncIterator[bytes],
    ) -> None:
        self._generator = generator

    def __aiter__(
        self,
    ) -> "AsyncStreamWrapper":
        return self

    async def __anext__(self) -> bytes:
        return await self._generator.__anext__()

    async def aclose(self) -> None:
        pass


def test_response_iter_bytes_no_encoding():
    body = [b"Hello, ", b"World!"]
    response = Response(
        status=200,
        headers=Headers(),
        content=StreamWrapper(iter(body)),
    )

    chunks = list(response.iter_bytes())
    result = b"".join(chunks)
    assert result == b"Hello, World!"


def test_response_iter_bytes_gzip():
    original = b"Hello, World! This is a test message for gzip compression."
    compressed = gzip.compress(original)

    response = Response(
        status=200,
        headers=Headers([("Content-Encoding", "gzip")]),
        content=StreamWrapper(iter([compressed])),
    )

    chunks = list(response.iter_bytes())
    result = b"".join(chunks)
    assert result == original


def test_response_iter_bytes_gzip_streaming():
    original = b"Hello, World! This is a test message for gzip compression."
    compressed = gzip.compress(original)

    chunk_size = 10
    body_chunks = [
        compressed[i : i + chunk_size]
        for i in range(
            0,
            len(compressed),
            chunk_size,
        )
    ]

    response = Response(
        status=200,
        headers=Headers([("Content-Encoding", "gzip")]),
        content=StreamWrapper(iter(body_chunks)),
    )

    chunks = list(response.iter_bytes())
    result = b"".join(chunks)
    assert result == original


def test_response_iter_bytes_deflate():
    original = b"Hello, World! This is a test message for deflate compression."
    compressed = zlib.compress(original)

    response = Response(
        status=200,
        headers=Headers(
            [
                (
                    "Content-Encoding",
                    "deflate",
                )
            ]
        ),
        content=StreamWrapper(iter([compressed])),
    )

    chunks = list(response.iter_bytes())
    result = b"".join(chunks)
    assert result == original


def test_response_iter_bytes_multiple_encodings():
    original = b"Hello, World! This is a test message for multi-layer compression."
    compressed_gzip = gzip.compress(original)
    compressed_deflate = zlib.compress(compressed_gzip)

    response = Response(
        status=200,
        headers=Headers(
            [
                (
                    "Content-Encoding",
                    "gzip, deflate",
                )
            ]
        ),
        content=StreamWrapper(iter([compressed_deflate])),
    )

    chunks = list(response.iter_bytes())
    result = b"".join(chunks)
    assert result == original


def test_response_iter_bytes_case_insensitive():
    original = b"Hello, World! This is a test message for gzip compression."
    compressed = gzip.compress(original)

    response = Response(
        status=200,
        headers=Headers([("Content-Encoding", "GZIP")]),
        content=StreamWrapper(iter([compressed])),
    )

    chunks = list(response.iter_bytes())
    result = b"".join(chunks)
    assert result == original


def test_response_iter_bytes_with_spaces():
    original = b"Hello, World! This is a test message for multi-layer compression."
    compressed_gzip = gzip.compress(original)
    compressed_deflate = zlib.compress(compressed_gzip)

    response = Response(
        status=200,
        headers=Headers(
            [
                (
                    "Content-Encoding",
                    " gzip , deflate ",
                )
            ]
        ),
        content=StreamWrapper(iter([compressed_deflate])),
    )

    chunks = list(response.iter_bytes())
    result = b"".join(chunks)
    assert result == original


def test_response_iter_bytes_unknown_encoding():
    original = b"Hello, World!"

    response = Response(
        status=200,
        headers=Headers(
            [
                (
                    "Content-Encoding",
                    "unknown",
                )
            ]
        ),
        content=StreamWrapper(iter([original])),
    )

    chunks = list(response.iter_bytes())
    result = b"".join(chunks)
    assert result == original


def test_response_iter_raw():
    original = b"Hello, World! This is a test message for gzip compression."
    compressed = gzip.compress(original)

    response = Response(
        status=200,
        headers=Headers([("Content-Encoding", "gzip")]),
        content=StreamWrapper(iter([compressed])),
    )

    chunks = list(response.iter_raw())
    result = b"".join(chunks)
    assert result == compressed
    assert result.startswith(b"\x1f\x8b")


def test_response_iter_raw_streaming():
    original = b"Hello, World! This is a test message for gzip compression."
    compressed = gzip.compress(original)

    chunk_size = 10
    body_chunks = [
        compressed[i : i + chunk_size]
        for i in range(
            0,
            len(compressed),
            chunk_size,
        )
    ]

    response = Response(
        status=200,
        headers=Headers([("Content-Encoding", "gzip")]),
        content=StreamWrapper(iter(body_chunks)),
    )

    chunks = list(response.iter_raw())
    result = b"".join(chunks)
    assert result == compressed


def test_response_read():
    original = b"Hello, World! This is a test message for gzip compression."
    compressed = gzip.compress(original)

    response = Response(
        status=200,
        headers=Headers([("Content-Encoding", "gzip")]),
        content=StreamWrapper(iter([compressed])),
    )

    content = response.read()
    assert content == original

    content2 = response.read()
    assert content2 == original
    assert content2 is content


def test_response_read_then_iter():
    original = b"Hello, World! This is a test message for gzip compression."
    compressed = gzip.compress(original)

    response = Response(
        status=200,
        headers=Headers([("Content-Encoding", "gzip")]),
        content=StreamWrapper(iter([compressed])),
    )

    content = response.read()
    assert content == original

    chunks = list(response.iter_bytes())
    result = b"".join(chunks)
    assert result == original


@pytest.mark.asyncio
async def test_response_async_iter_bytes_gzip():
    original = b"Hello, World! This is a test message for gzip compression."
    compressed = gzip.compress(original)

    async def body_generator():
        yield compressed

    response = Response(
        status=200,
        headers=Headers([("Content-Encoding", "gzip")]),
        content=AsyncStreamWrapper(body_generator()),
    )

    chunks = [chunk async for chunk in response.async_iter_bytes()]
    result = b"".join(chunks)
    assert result == original


@pytest.mark.asyncio
async def test_response_async_iter_bytes_streaming():
    original = b"Hello, World! This is a test message for gzip compression."
    compressed = gzip.compress(original)

    chunk_size = 10
    body_chunks = [
        compressed[i : i + chunk_size]
        for i in range(
            0,
            len(compressed),
            chunk_size,
        )
    ]

    async def body_generator():
        for chunk in body_chunks:
            yield chunk

    response = Response(
        status=200,
        headers=Headers([("Content-Encoding", "gzip")]),
        content=AsyncStreamWrapper(body_generator()),
    )

    chunks = [chunk async for chunk in response.async_iter_bytes()]
    result = b"".join(chunks)
    assert result == original


@pytest.mark.asyncio
async def test_response_async_iter_raw():
    original = b"Hello, World! This is a test message for gzip compression."
    compressed = gzip.compress(original)

    async def body_generator():
        yield compressed

    response = Response(
        status=200,
        headers=Headers([("Content-Encoding", "gzip")]),
        content=AsyncStreamWrapper(body_generator()),
    )

    chunks = [chunk async for chunk in response.async_iter_raw()]
    result = b"".join(chunks)
    assert result == compressed
    assert result.startswith(b"\x1f\x8b")


@pytest.mark.asyncio
async def test_response_aread():
    original = b"Hello, World! This is a test message for gzip compression."
    compressed = gzip.compress(original)

    async def body_generator():
        yield compressed

    response = Response(
        status=200,
        headers=Headers([("Content-Encoding", "gzip")]),
        content=AsyncStreamWrapper(body_generator()),
    )

    content = await response.aread()
    assert content == original

    content2 = await response.aread()
    assert content2 == original
    assert content2 is content


@pytest.mark.asyncio
async def test_response_aread_then_async_iter():
    original = b"Hello, World! This is a test message for gzip compression."
    compressed = gzip.compress(original)

    async def body_generator():
        yield compressed

    response = Response(
        status=200,
        headers=Headers([("Content-Encoding", "gzip")]),
        content=AsyncStreamWrapper(body_generator()),
    )

    content = await response.aread()
    assert content == original

    chunks = [chunk async for chunk in response.async_iter_bytes()]
    result = b"".join(chunks)
    assert result == original


def test_response_iter_bytes_custom_chunk_size():
    original = b"Hello, World! This is a test message for gzip compression."
    compressed = gzip.compress(original)

    response = Response(
        status=200,
        headers=Headers([("Content-Encoding", "gzip")]),
        content=StreamWrapper(iter([compressed])),
    )

    chunks = list(response.iter_bytes(chunk_size=10))
    assert all(len(chunk) <= 10 for chunk in chunks[:-1])

    result = b"".join(chunks)
    assert result == original


def test_response_brotli_not_installed():
    has_brotli = importlib.util.find_spec("brotli") is not None
    has_brotlicffi = importlib.util.find_spec("brotlicffi") is not None
    if has_brotli:
        pytest.skip("brotli is installed")
    if has_brotlicffi:
        pytest.skip("brotlicffi is installed")

    response = Response(
        status=200,
        headers=Headers([("Content-Encoding", "br")]),
        content=StreamWrapper(iter([b"some data"])),
    )

    chunks = list(response.iter_bytes())
    result = b"".join(chunks)
    assert result == b"some data"


def test_response_zstd_not_installed():
    has_zstandard = importlib.util.find_spec("zstandard") is not None
    if has_zstandard:
        pytest.skip("zstandard is installed")

    response = Response(
        status=200,
        headers=Headers([("Content-Encoding", "zstd")]),
        content=StreamWrapper(iter([b"some data"])),
    )

    chunks = list(response.iter_bytes())
    result = b"".join(chunks)
    assert result == b"some data"


def test_response_brotli_if_installed():
    try:
        import brotli
    except ImportError:
        try:
            import brotlicffi as brotli  # type: ignore[import-not-found, no-redef]
        except ImportError:
            pytest.skip("brotli not installed")

    original = b"Hello, World! This is a test message for brotli compression."
    compressed = brotli.compress(original)  # type: ignore[attr-defined]

    response = Response(
        status=200,
        headers=Headers([("Content-Encoding", "br")]),
        content=StreamWrapper(iter([compressed])),
    )

    chunks = list(response.iter_bytes())
    result = b"".join(chunks)
    assert result == original


def test_response_zstd_if_installed():
    try:
        import zstandard
    except ImportError:
        pytest.skip("zstandard not installed")

    original = b"Hello, World! This is a test message for zstd compression."
    compressor = zstandard.ZstdCompressor()
    compressed = compressor.compress(original)

    response = Response(
        status=200,
        headers=Headers([("Content-Encoding", "zstd")]),
        content=StreamWrapper(iter([compressed])),
    )

    chunks = list(response.iter_bytes())
    result = b"".join(chunks)
    assert result == original


def test_response_identity_encoding():
    original = b"Hello, World!"

    response = Response(
        status=200,
        headers=Headers(
            [
                (
                    "Content-Encoding",
                    "identity",
                )
            ]
        ),
        content=StreamWrapper(iter([original])),
    )

    chunks = list(response.iter_bytes())
    result = b"".join(chunks)
    assert result == original


def test_response_empty_body():
    response = Response(
        status=204,
        headers=Headers(),
        content=None,
    )

    chunks = list(response.iter_bytes())
    assert chunks == []


def test_response_empty_body_with_encoding():
    response = Response(
        status=204,
        headers=Headers([("Content-Encoding", "gzip")]),
        content=None,
    )

    chunks = list(response.iter_bytes())
    assert chunks == []



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from zapros._base_pool import PoolKey
    from zapros._models import Request

DEFAULT_PORTS = {
    "http": "80",
    "https": "443",
}


def get_host_header_value(host: str, scheme: str, port: str) -> str:
    default_port = DEFAULT_PORTS.get(scheme.lower())
    if port == "" or port == default_port:
        return host
    return f"{host}:{port}"


def get_authority_value(host: str, port: str) -> str:
    if port == "":
        return host
    return f"{host}:{port}"


def get_pool_key(
    request: Request,
    target_scheme: str,
    target_host: str,
    target_port: int,
) -> tuple[PoolKey, bool]:
    proxy_context = request.context.get("network", {}).get("proxy")
    proxy_url_value = proxy_context.get("url") if proxy_context else None

    if proxy_url_value is None:
        return ((target_scheme, target_host, target_port), False)

    from pywhatwgurl import URL

    proxy_url = URL(proxy_url_value) if isinstance(proxy_url_value, str) else proxy_url_value
    proxy_scheme = proxy_url.protocol[:-1]
    proxy_host = proxy_url.hostname
    proxy_port = int(proxy_url.port) if proxy_url.port != "" else (443 if proxy_scheme == "https" else 80)

    is_target_https = target_scheme in ("https", "wss")

    if is_target_https:
        return ((proxy_scheme, proxy_host, proxy_port, target_scheme, target_host, target_port), False)
    else:
        return ((proxy_scheme, proxy_host, proxy_port), True)



from typing import (
    TYPE_CHECKING,
    AsyncIterable,
    AsyncIterator,
    Iterable,
    Iterator,
    Union,
    cast,
    overload,
)

import typing_extensions

if TYPE_CHECKING:
    from hishel import (
        AsyncBaseStorage as HishelAsyncBaseStorage,
        AsyncCacheProxy,
        CachePolicy,
        Headers,
        Request as HishelRequest,
        Response as HishelResponse,
        SyncBaseStorage as HishelSyncBaseStorage,
        SyncCacheProxy,
    )
else:
    try:
        from hishel import (
            AsyncBaseStorage as HishelAsyncBaseStorage,
            AsyncCacheProxy,
            CachePolicy,
            Headers,
            Request as HishelRequest,
            Response as HishelResponse,
            SyncBaseStorage as HishelSyncBaseStorage,
            SyncCacheProxy,
        )
    except ImportError:
        HishelAsyncBaseStorage = None
        AsyncCacheProxy = None
        CachePolicy = None
        Headers = None
        HishelRequest = None
        HishelResponse = None
        HishelSyncBaseStorage = None
        SyncCacheProxy = None
from pywhatwgurl import URL

from zapros._models import (
    Request,
    Response,
    ResponseCachingContext,
)

from ._async_base import (
    AsyncBaseHandler,
    AsyncBaseMiddleware,
)
from ._sync_base import (
    BaseHandler,
    BaseMiddleware,
)


class _BytesStream(
    AsyncIterable[bytes],
    Iterable[bytes],
):
    def __init__(self, data: bytes) -> None:
        self._data = data
        self._sent = False

    def __iter__(
        self,
    ) -> "_BytesStream":
        return self

    def __next__(self) -> bytes:
        if self._sent:
            raise StopIteration
        self._sent = True
        return self._data

    def __aiter__(
        self,
    ) -> "_BytesStream":
        return self

    async def __anext__(self) -> bytes:
        if self._sent:
            raise StopAsyncIteration
        self._sent = True
        return self._data


@overload
def _zapros_to_hishel(
    model: Request,
) -> "HishelRequest": ...
@overload
def _zapros_to_hishel(
    model: Response,
) -> "HishelResponse": ...
def _zapros_to_hishel(
    model: Request | Response,
) -> Union["HishelRequest", "HishelResponse"]:

    if isinstance(model, Request):
        stream: Iterator[bytes] | AsyncIterator[bytes]

        if model.body is None:
            stream = _BytesStream(b"")
        elif isinstance(model.body, bytes):
            stream = _BytesStream(model.body)
        else:
            stream = model.body
        return HishelRequest(
            method=model.method,
            url=str(model.url.to_string()),
            headers=Headers(dict(model.headers.list())),
            stream=stream,
            metadata={"hishel_" + key: value for key, value in model.context.get("caching", {}).items()},
        )
    else:
        stream: Iterator[bytes] | AsyncIterator[bytes]

        if model.content is None:
            stream = _BytesStream(b"")
        elif isinstance(model.content, bytes):
            stream = _BytesStream(model.content)
        else:
            stream = model.content

        return HishelResponse(
            status_code=model.status,
            headers=Headers(dict(model.headers.list())),
            stream=stream,
            metadata={},
        )


@overload
def _hishel_to_zapros(
    model: "HishelRequest",
) -> Request: ...
@overload
def _hishel_to_zapros(
    model: "HishelResponse",
) -> Response: ...
def _hishel_to_zapros(
    model: Union[HishelRequest, HishelResponse],
) -> Request | Response:
    if isinstance(model, HishelRequest):
        return Request(
            method=model.method,
            url=URL(model.url),
            headers=Headers(dict(model.headers)),
            body=model.stream,
            context={
                "caching": {
                    key.removeprefix("hishel_"): value
                    for key, value in model.metadata.items()
                    if key.startswith("hishel_")
                }  # type: ignore
            },
        )
    else:
        context: ResponseCachingContext = {}

        if "hishel_from_cache" in model.metadata:
            context["from_cache"] = model.metadata["hishel_from_cache"]
        if "hishel_created_at" in model.metadata:
            context["created_at"] = model.metadata["hishel_created_at"]
        if "hishel_revalidated" in model.metadata:
            context["revalidated"] = model.metadata["hishel_revalidated"]
        if "hishel_stored" in model.metadata:
            context["stored"] = model.metadata["hishel_stored"]

        return Response(
            status=model.status_code,
            headers=Headers(dict(model.headers)),
            content=model.stream,
            context={"caching": context},
        )


class CacheMiddleware(AsyncBaseMiddleware, BaseMiddleware):
    @overload
    def __init__(
        self,
        next_handler: AsyncBaseHandler,
        *,
        storage: Union[HishelAsyncBaseStorage, None] = None,
        policy: Union[CachePolicy, None] = None,
    ) -> None: ...
    @overload
    def __init__(
        self,
        next_handler: BaseHandler,
        *,
        storage: Union[HishelSyncBaseStorage, None] = None,
        policy: Union[CachePolicy, None] = None,
    ) -> None: ...
    def __init__(
        self,
        next_handler: AsyncBaseHandler | BaseHandler,
        *,
        storage: Union[HishelSyncBaseStorage, HishelAsyncBaseStorage, None] = None,
        policy: Union[CachePolicy, None] = None,
    ) -> None:

        if CachePolicy is None:  # type: ignore[reportUnboundVariable]
            raise ImportError(
                "hishel is not installed. Install zapros with caching feature: pip install zapros[caching]"
            )

        self.next = cast(BaseMiddleware, next_handler)
        self.async_next = cast(
            AsyncBaseMiddleware,
            next_handler,
        )
        self._storage = storage
        self._policy = policy
        self._cache_proxy: AsyncCacheProxy | SyncCacheProxy | None = None

    def _get_sync_cache_proxy(
        self,
    ) -> SyncCacheProxy:
        if self._cache_proxy is None:

            def _sync_send_request(
                request: HishelRequest,
            ) -> HishelResponse:
                zapros_request = _hishel_to_zapros(request)
                zapros_response = self.next.handle(zapros_request)
                return _zapros_to_hishel(zapros_response)

            assert (
                isinstance(
                    self._storage,
                    HishelSyncBaseStorage,
                )
                or self._storage is None
            ), "Sync handler requires sync storage"
            self._cache_proxy = SyncCacheProxy(
                request_sender=_sync_send_request,
                storage=self._storage,
                policy=self._policy,
            )

        if not isinstance(
            self._cache_proxy,
            SyncCacheProxy,
        ):
            raise TypeError("Cache proxy is not a SyncCacheProxy")
        return self._cache_proxy

    async def _get_async_cache_proxy(
        self,
    ) -> AsyncCacheProxy:
        if self._cache_proxy is None:

            async def _async_send_request(
                request: HishelRequest,
            ) -> HishelResponse:
                zapros_request = _hishel_to_zapros(request)
                zapros_response = await self.async_next.ahandle(zapros_request)
                return _zapros_to_hishel(zapros_response)

            assert (
                isinstance(
                    self._storage,
                    HishelAsyncBaseStorage,
                )
                or self._storage is None
            ), "Async handler requires async storage"
            self._cache_proxy = AsyncCacheProxy(
                request_sender=_async_send_request,
                storage=self._storage,
                policy=self._policy,
            )

        if not isinstance(
            self._cache_proxy,
            AsyncCacheProxy,
        ):
            raise TypeError("Cache proxy is not an AsyncCacheProxy")
        return self._cache_proxy

    async def ahandle(self, request: Request) -> Response:
        hishel_request = _zapros_to_hishel(request)
        proxy = await self._get_async_cache_proxy()
        hishel_response = await proxy.handle_request(hishel_request)
        return _hishel_to_zapros(hishel_response)

    def handle(self, request: Request) -> Response:
        hishel_request = _zapros_to_hishel(request)
        hishel_response = self._get_sync_cache_proxy().handle_request(hishel_request)
        return _hishel_to_zapros(hishel_response)


@typing_extensions.deprecated(
    "CachingHandler is deprecated, use CacheMiddleware instead. "
    "The name 'Handler' was misleading as this is a middleware, not a terminal handler."
)
class CachingHandler(CacheMiddleware):
    pass



import base64

import pytest

from zapros import AsyncClient, Client
from zapros._handlers._mock import (
    Mock,
    MockMiddleware,
    MockRouter,
)
from zapros._models import Response
from zapros.matchers import path


def test_bearer_token_sync():
    router = MockRouter()
    Mock.given(
        path("/api").header(
            "authorization",
            "Bearer test-token",
        )
    ).respond(Response(status=200)).mount(router)

    with Client(
        handler=MockMiddleware(router),
        auth="test-token",
    ) as client:
        response = client.get(
            "https://api.example.com/api",
        )
        assert response.status == 200


@pytest.mark.asyncio
async def test_bearer_token_async():
    router = MockRouter()
    Mock.given(
        path("/api").header(
            "authorization",
            "Bearer test-token",
        )
    ).respond(Response(status=200)).mount(router)

    async with AsyncClient(
        handler=MockMiddleware(router),
        auth="test-token",
    ) as client:
        response = await client.get(
            "https://api.example.com/api",
        )
        assert response.status == 200


def test_basic_auth_sync():
    router = MockRouter()
    username = "user"
    password = "pass"
    expected_credentials = base64.b64encode(f"{username}:{password}".encode()).decode()
    expected_header = f"Basic {expected_credentials}"

    Mock.given(
        path("/api").header(
            "authorization",
            expected_header,
        )
    ).respond(Response(status=200)).mount(router)

    with Client(
        handler=MockMiddleware(router),
        auth=(username, password),
    ) as client:
        response = client.get(
            "https://api.example.com/api",
        )
        assert response.status == 200


@pytest.mark.asyncio
async def test_basic_auth_async():
    router = MockRouter()
    username = "user"
    password = "pass"
    expected_credentials = base64.b64encode(f"{username}:{password}".encode()).decode()
    expected_header = f"Basic {expected_credentials}"

    Mock.given(
        path("/api").header(
            "authorization",
            expected_header,
        )
    ).respond(Response(status=200)).mount(router)

    async with AsyncClient(
        handler=MockMiddleware(router),
        auth=(username, password),
    ) as client:
        response = await client.get(
            "https://api.example.com/api",
        )
        assert response.status == 200


def test_no_auth_sync():
    router = MockRouter()
    Mock.given(path("/api")).respond(Response(status=200)).mount(router)

    with Client(handler=MockMiddleware(router)) as client:
        response = client.get(
            "https://api.example.com/api",
        )
        assert response.status == 200


@pytest.mark.asyncio
async def test_no_auth_async():
    router = MockRouter()
    Mock.given(path("/api")).respond(Response(status=200)).mount(router)

    async with AsyncClient(handler=MockMiddleware(router)) as client:
        response = await client.get(
            "https://api.example.com/api",
        )
        assert response.status == 200


def test_bearer_with_default_headers_sync():
    router = MockRouter()
    Mock.given(
        path("/api")
        .header(
            "authorization",
            "Bearer token123",
        )
        .header("x-custom", "value")
    ).respond(Response(status=200)).mount(router)

    with Client(
        handler=MockMiddleware(router),
        auth="token123",
        default_headers={"x-custom": "value"},
    ) as client:
        response = client.get(
            "https://api.example.com/api",
        )
        assert response.status == 200


@pytest.mark.asyncio
async def test_bearer_with_default_headers_async():
    router = MockRouter()
    Mock.given(
        path("/api")
        .header(
            "authorization",
            "Bearer token123",
        )
        .header("x-custom", "value")
    ).respond(Response(status=200)).mount(router)

    async with AsyncClient(
        handler=MockMiddleware(router),
        auth="token123",
        default_headers={"x-custom": "value"},
    ) as client:
        response = await client.get(
            "https://api.example.com/api",
        )
        assert response.status == 200


def test_with_handler_preserves_bearer_sync():
    router = MockRouter()
    Mock.given(
        path("/api").header(
            "authorization",
            "Bearer my-token",
        )
    ).respond(Response(status=200)).mount(router)

    client = Client(
        handler=MockMiddleware(router),
        auth="my-token",
    )

    response = client.get(
        "https://api.example.com/api",
    )
    assert response.status == 200

    client.close()


@pytest.mark.asyncio
async def test_with_handler_preserves_bearer_async():
    router = MockRouter()
    Mock.given(
        path("/api").header(
            "authorization",
            "Bearer my-token",
        )
    ).respond(Response(status=200)).mount(router)

    client = AsyncClient(
        handler=MockMiddleware(router),
        auth="my-token",
    )

    response = await client.get(
        "https://api.example.com/api",
    )
    assert response.status == 200

    await client.aclose()


def test_with_handler_preserves_basic_auth_sync():
    router = MockRouter()
    username = "user"
    password = "pass"
    expected_credentials = base64.b64encode(f"{username}:{password}".encode()).decode()
    expected_header = f"Basic {expected_credentials}"

    Mock.given(
        path("/api").header(
            "authorization",
            expected_header,
        )
    ).respond(Response(status=200)).mount(router)

    client = Client(
        handler=MockMiddleware(router),
        auth=(username, password),
    )

    response = client.get(
        "https://api.example.com/api",
    )
    assert response.status == 200

    client.close()


@pytest.mark.asyncio
async def test_with_handler_preserves_basic_auth_async():
    router = MockRouter()
    username = "user"
    password = "pass"
    expected_credentials = base64.b64encode(f"{username}:{password}".encode()).decode()
    expected_header = f"Basic {expected_credentials}"

    Mock.given(
        path("/api").header(
            "authorization",
            expected_header,
        )
    ).respond(Response(status=200)).mount(router)

    client = AsyncClient(
        handler=MockMiddleware(router),
        auth=(username, password),
    )

    response = await client.get(
        "https://api.example.com/api",
    )
    assert response.status == 200

    await client.aclose()


def test_per_request_bearer_sync():
    router = MockRouter()
    Mock.given(
        path("/api").header(
            "authorization",
            "Bearer request-token",
        )
    ).respond(Response(status=200)).mount(router)

    with Client(handler=MockMiddleware(router)) as client:
        response = client.get(
            "https://api.example.com/api",
            auth="request-token",
        )
        assert response.status == 200


@pytest.mark.asyncio
async def test_per_request_bearer_async():
    router = MockRouter()
    Mock.given(
        path("/api").header(
            "authorization",
            "Bearer request-token",
        )
    ).respond(Response(status=200)).mount(router)

    async with AsyncClient(handler=MockMiddleware(router)) as client:
        response = await client.get(
            "https://api.example.com/api",
            auth="request-token",
        )
        assert response.status == 200


def test_per_request_basic_auth_sync():
    router = MockRouter()
    username = "req_user"
    password = "req_pass"
    expected_credentials = base64.b64encode(f"{username}:{password}".encode()).decode()
    expected_header = f"Basic {expected_credentials}"

    Mock.given(
        path("/api").header(
            "authorization",
            expected_header,
        )
    ).respond(Response(status=200)).mount(router)

    with Client(handler=MockMiddleware(router)) as client:
        response = client.get(
            "https://api.example.com/api",
            auth=(username, password),
        )
        assert response.status == 200


@pytest.mark.asyncio
async def test_per_request_basic_auth_async():
    router = MockRouter()
    username = "req_user"
    password = "req_pass"
    expected_credentials = base64.b64encode(f"{username}:{password}".encode()).decode()
    expected_header = f"Basic {expected_credentials}"

    Mock.given(
        path("/api").header(
            "authorization",
            expected_header,
        )
    ).respond(Response(status=200)).mount(router)

    async with AsyncClient(handler=MockMiddleware(router)) as client:
        response = await client.get(
            "https://api.example.com/api",
            auth=(username, password),
        )
        assert response.status == 200


def test_per_request_overrides_client_bearer_sync():
    router = MockRouter()
    Mock.given(
        path("/api").header(
            "authorization",
            "Bearer request-token",
        )
    ).respond(Response(status=200)).mount(router)

    with Client(
        handler=MockMiddleware(router),
        auth="client-token",
    ) as client:
        response = client.get(
            "https://api.example.com/api",
            auth="request-token",
        )
        assert response.status == 200


@pytest.mark.asyncio
async def test_per_request_overrides_client_bearer_async():
    router = MockRouter()
    Mock.given(
        path("/api").header(
            "authorization",
            "Bearer request-token",
        )
    ).respond(Response(status=200)).mount(router)

    async with AsyncClient(
        handler=MockMiddleware(router),
        auth="client-token",
    ) as client:
        response = await client.get(
            "https://api.example.com/api",
            auth="request-token",
        )
        assert response.status == 200


def test_per_request_overrides_client_basic_auth_sync():
    router = MockRouter()
    username = "req_user"
    password = "req_pass"
    expected_credentials = base64.b64encode(f"{username}:{password}".encode()).decode()
    expected_header = f"Basic {expected_credentials}"

    Mock.given(
        path("/api").header(
            "authorization",
            expected_header,
        )
    ).respond(Response(status=200)).mount(router)

    with Client(
        handler=MockMiddleware(router),
        auth=(
            "client_user",
            "client_pass",
        ),
    ) as client:
        response = client.get(
            "https://api.example.com/api",
            auth=(username, password),
        )
        assert response.status == 200


@pytest.mark.asyncio
async def test_per_request_overrides_client_basic_auth_async():
    router = MockRouter()
    username = "req_user"
    password = "req_pass"
    expected_credentials = base64.b64encode(f"{username}:{password}".encode()).decode()
    expected_header = f"Basic {expected_credentials}"

    Mock.given(
        path("/api").header(
            "authorization",
            expected_header,
        )
    ).respond(Response(status=200)).mount(router)

    async with AsyncClient(
        handler=MockMiddleware(router),
        auth=(
            "client_user",
            "client_pass",
        ),
    ) as client:
        response = await client.get(
            "https://api.example.com/api",
            auth=(username, password),
        )
        assert response.status == 200



import pytest

from zapros import AsyncClient
from zapros._handlers._mock import Mock
from zapros._models import (
    Headers,
    Request,
    Response,
)
from zapros.matchers import method, path
from zapros.mock import mock_http


@pytest.mark.asyncio
async def test_verification_on_exit():
    with pytest.raises(
        AssertionError,
        match="expected 1 calls, got 0",
    ):
        with mock_http() as router:
            Mock().given(path("/api")).respond(Response(status=200)).once().mount(router)


@pytest.mark.asyncio
async def test_no_verification_on_exception():
    try:
        with mock_http() as router:
            Mock().given(path("/api")).respond(Response(status=200)).once().mount(router)
            raise RuntimeError("test error")
    except RuntimeError:
        pass


@pytest.mark.asyncio
async def test_multiple_mocks():
    with mock_http() as router:
        Mock().given(path("/users")).respond(Response(status=200, json=[{"id": 1}])).mount(router)
        Mock().given(path("/posts").method("POST")).respond(Response(status=201, json={"id": 1})).mount(router)
        Mock().given(path("/health")).respond(Response(status=200, text="OK")).mount(router)

        async with AsyncClient() as client:
            response = await client.get(
                "https://api.example.com/users",
            )
            assert response.status == 200
            assert response.json == [{"id": 1}]

            response = await client.post(
                "https://api.example.com/posts",
                json={"title": "Test"},
            )
            assert response.status == 201
            assert response.json == {"id": 1}

            response = await client.get(
                "https://api.example.com/health",
            )
            assert response.status == 200
            assert response.text == "OK"


@pytest.mark.asyncio
async def test_context_manager_with_headers():
    with mock_http() as router:
        Mock().given(path("/api").header("x-api-key", "secret")).respond(
            Response(
                status=200,
                json={"authenticated": True},
            )
        ).mount(router)

        async with AsyncClient() as client:
            headers = Headers({"x-api-key": "secret"})
            response = await client.get(
                "https://api.example.com/api",
                headers=headers,
            )
            assert response.status == 200
            assert response.json == {"authenticated": True}


@pytest.mark.asyncio
async def test_context_manager_with_query():
    with mock_http() as router:
        Mock().given(path("/search").query(q="test", page="1")).respond(
            Response(
                status=200,
                json={"results": []},
            )
        ).mount(router)

        async with AsyncClient() as client:
            response = await client.get(
                "https://api.example.com/search?q=test&page=1",
            )
            assert response.status == 200
            assert response.json == {"results": []}


@pytest.mark.asyncio
async def test_context_manager_first_match_wins():
    with mock_http() as router:
        Mock().given(path("/api")).respond(Response(status=200, text="First")).mount(router)
        Mock().given(path("/api")).respond(Response(status=200, text="Second")).mount(router)

        async with AsyncClient() as client:
            response = await client.get(
                "https://api.example.com/api",
            )
            assert response.status == 200
            assert response.text == "First"


@pytest.mark.asyncio
async def test_callback_in_context():
    def custom_handler(req: Request):
        from zapros._models import (
            Response,
        )

        status = 404 if req.url.pathname == "/notfound" else 200
        return Response(
            status=status,
            headers=Headers({}),
            content=None,
        )

    with mock_http() as router:
        Mock().given(method("GET")).callback(custom_handler).mount(router)

        async with AsyncClient() as client:
            response = await client.get(
                "https://api.example.com/found",
            )
            assert response.status == 200

            response = await client.get(
                "https://api.example.com/notfound",
            )
            assert response.status == 404


@pytest.mark.asyncio
async def test_expect_verification():
    with mock_http() as router:
        Mock().given(path("/api")).respond(Response(status=200)).expect(2).mount(router)

        async with AsyncClient() as client:
            await client.get(
                "https://api.example.com/api",
            )
            await client.get(
                "https://api.example.com/api",
            )


@pytest.mark.asyncio
async def test_expect_verification_failure():
    with pytest.raises(
        AssertionError,
        match="API Mock: expected 3 calls, got 2",
    ):
        with mock_http() as router:
            Mock().given(path("/api")).respond(Response(status=200)).expect(3).mount(router).name("API Mock")

            async with AsyncClient() as client:
                await client.get(
                    "https://api.example.com/api",
                )
                await client.get(
                    "https://api.example.com/api",
                )


@pytest.mark.asyncio
async def test_never_expectation():
    with mock_http() as router:
        Mock().given(path("/api")).respond(Response(status=200)).never().mount(router)


@pytest.mark.asyncio
async def test_never_expectation_failure():
    with pytest.raises(
        AssertionError,
        match="expected 0 calls, got 1",
    ):
        with mock_http() as router:
            Mock().given(path("/api")).respond(Response(status=200)).never().mount(router)

            async with AsyncClient() as client:
                await client.get(
                    "https://api.example.com/api",
                )


@pytest.mark.asyncio
async def test_different_methods():
    with mock_http() as router:
        Mock().given(path("/api").method("GET")).respond(
            Response(
                status=200,
                json={"method": "GET"},
            )
        ).mount(router)
        Mock().given(path("/api").method("POST")).respond(
            Response(
                status=201,
                json={"method": "POST"},
            )
        ).mount(router)
        Mock().given(path("/api").method("PUT")).respond(
            Response(
                status=200,
                json={"method": "PUT"},
            )
        ).mount(router)
        Mock().given(path("/api").method("DELETE")).respond(Response(status=204)).mount(router)

        async with AsyncClient() as client:
            response = await client.get(
                "https://api.example.com/api",
            )
            assert response.status == 200
            assert response.json == {"method": "GET"}

            response = await client.post(
                "https://api.example.com/api",
                json={},
            )
            assert response.status == 201
            assert response.json == {"method": "POST"}

            response = await client.put(
                "https://api.example.com/api",
                json={},
            )
            assert response.status == 200
            assert response.json == {"method": "PUT"}

            response = await client.request(
                "DELETE",
                "https://api.example.com/api",
            )
            assert response.status == 204


@pytest.mark.asyncio
async def test_different_hosts():
    with mock_http() as router:
        from zapros.matchers import host

        Mock().given(host("api1.example.com").path("/data")).respond(
            Response(
                status=200,
                json={"host": "api1"},
            )
        ).mount(router)
        Mock().given(host("api2.example.com").path("/data")).respond(
            Response(
                status=200,
                json={"host": "api2"},
            )
        ).mount(router)

        async with AsyncClient() as client:
            response = await client.get(
                "https://api1.example.com/data",
            )
            assert response.status == 200
            assert response.json == {"host": "api1"}

            response = await client.get(
                "https://api2.example.com/data",
            )
            assert response.status == 200
            assert response.json == {"host": "api2"}



from __future__ import annotations

import base64
import ssl
import time
import warnings
from collections.abc import (
    Iterator as ABCIterator,
)
from typing import cast

import h11
from pywhatwgurl import URL
from typing_extensions import override

from zapros._constants import DEFAULT_READ_SIZE, DEFAULT_SSL_CONTEXT
from zapros._io._sync import SyncTransport
from zapros._io._base import BaseNetworkStream, BaseTransport
from zapros._utils import get_authority_value, get_pool_key

from .._sync_pool import ConnPool
from .._base_pool import PoolKey
from .._errors import (
    ConnectionError,
    TotalTimeoutError,
)
from .._models import (
    ClosableStream,
    Request,
    Response,
    ResponseHandoffContext,
)
from ._sync_base import BaseHandler
from ._exc_map import map_read_exceptions, map_write_exceptions


def _encode_target(path: str, query: str) -> bytes:
    """Encode an already-prepared request target without rewriting semantics.

    The request object is expected to contain a fully prepared / already-encoded
    path and query. We only join them and convert to ASCII bytes for h11.
    """
    prepared_path = path or "/"
    if query:
        return f"{prepared_path}?{query}".encode("ascii")
    return prepared_path.encode("ascii")


def _header_has_token(
    headers: list[tuple[str, str]],
    name: str,
    token: str,
) -> bool:
    name = name.lower()
    token = token.lower()
    for k, v in headers:
        if k.lower() != name:
            continue
        for part in v.split(","):
            if part.strip().lower() == token:
                return True
    return False


def _min_timeout(a: float | None, b: float | None) -> float | None:
    if a is None:
        return b
    if b is None:
        return a
    return min(a, b)


class Conn:
    def __init__(
        self,
        stream: BaseNetworkStream,
    ) -> None:
        self.stream = stream
        self.h11 = h11.Connection(h11.CLIENT)
        self._closed = False

    def close(self) -> None:
        if self._closed:
            return
        self._closed = True
        try:
            self.stream.close()
        except Exception:
            pass

    @property
    def is_closed(self) -> bool:
        return self._closed

    def can_reuse(self) -> bool:
        return not self.is_closed and self.h11.our_state is h11.IDLE and self.h11.their_state is h11.IDLE


class StdStream(ClosableStream):
    """Wrap an h11 response stream and return the connection to the pool on close."""

    def __init__(
        self,
        conn: Conn,
        pool: ConnPool,
        key: PoolKey,
        *,
        read_timeout: float | None = None,
        no_body_response: bool = False,
        must_close: bool = False,
    ) -> None:
        self._conn = conn
        self._pool = pool
        self._key = key
        self._read_timeout = read_timeout
        self._closed = False
        self._eof = False
        self._ok = True
        self._no_body_response = no_body_response
        self._must_close = must_close

    def __iter__(self) -> StdStream:
        return self

    def _read(self, n: int) -> bytes:
        with map_read_exceptions():
            return self._conn.stream.read(n, timeout=self._read_timeout)

    def _drain_no_body_end_of_message(self) -> None:
        if self._eof or not self._no_body_response:
            return

        while True:
            event = self._conn.h11.next_event()
            if isinstance(event, h11.EndOfMessage):
                self._eof = True
                self._conn.h11.start_next_cycle()
                return
            if event is h11.NEED_DATA:
                self._ok = False
                return
            if isinstance(event, h11.Data):
                self._ok = False
                return

    def __next__(self) -> bytes:
        if self._closed:
            raise StopIteration

        try:
            while True:
                event = self._conn.h11.next_event()

                if event is h11.NEED_DATA:
                    data = self._read(DEFAULT_READ_SIZE)
                    if not data:
                        self._conn.h11.receive_data(b"")
                        continue
                    self._conn.h11.receive_data(data)
                    continue

                if isinstance(event, h11.Data):
                    return bytes(event.data)

                if isinstance(event, h11.EndOfMessage):
                    self._eof = True
                    try:
                        self._conn.h11.start_next_cycle()
                    except Exception:
                        self._ok = False
                    self.close()
                    raise StopIteration

        except BaseException:
            self._ok = False
            self.close()
            raise

    @override
    def close(self) -> None:
        if self._closed:
            return
        self._closed = True

        if self._no_body_response and not self._eof and self._ok:
            try:
                self._drain_no_body_end_of_message()
            except Exception:
                self._ok = False

        reuse = self._eof and self._ok and not self._must_close and self._conn.can_reuse()
        self._pool.release(self._key, self._conn, reuse=reuse)


class StdNetworkHandler(BaseHandler):
    def __init__(
        self,
        *,
        transport: BaseTransport | None = None,
        ssl_context: ssl.SSLContext | None = None,
        total_timeout: float | None = None,
        connect_timeout: float | None = None,
        read_timeout: float | None = None,
        write_timeout: float | None = None,
        max_connections_per_host: int = 10,
        max_idle_connections_per_host: int | None = None,
        max_idle_seconds: float = 30.0,
    ) -> None:
        if ssl_context is not None:
            warnings.warn(
                "The ssl_context argument is deprecated; set it through the transport argument instead",
                DeprecationWarning,
                stacklevel=2,
            )
        self.ssl_context = ssl_context or DEFAULT_SSL_CONTEXT

        self.transport = transport or SyncTransport(ssl_context=self.ssl_context)

        # Total timeout means: start of handle() until response headers received.
        self.total_timeout = total_timeout

        # Per-phase timeouts remain independently configurable.
        self.connect_timeout = connect_timeout if connect_timeout is not None else None
        self.read_timeout = read_timeout if read_timeout is not None else None
        self.write_timeout = write_timeout if write_timeout is not None else None

        self._pool = ConnPool(
            max_connections_per_host=max_connections_per_host,
            max_idle_per_host=(
                max_connections_per_host if max_idle_connections_per_host is None else max_idle_connections_per_host
            ),
            max_idle_seconds=max_idle_seconds,
        )

    @staticmethod
    def _remaining_timeout(deadline: float | None) -> float | None:
        if deadline is None:
            return None
        remaining = deadline - time.monotonic()
        if remaining <= 0:
            raise TotalTimeoutError("Operation timed out")
        return remaining

    def _new_conn(
        self,
        request: Request,
        scheme: str,
        host: str,
        port: int,
        *,
        connect_timeout: float | None = None,
    ) -> Conn:
        is_secure = scheme in ("https", "wss")
        proxy_context = request.context.get("network", {}).get("proxy")
        proxy_url = proxy_context.get("url") if proxy_context is not None else None

        if proxy_url is not None:
            proxy_url = URL(proxy_url) if isinstance(proxy_url, str) else proxy_url
            connect_host = proxy_url.hostname
            connect_port = int(proxy_url.port) if proxy_url.port != "" else 80
            use_tls = proxy_url.protocol in ("https:", "wss:")
        else:
            connect_host = host
            connect_port = port
            use_tls = is_secure

        stream = self.transport.connect(
            connect_host,
            connect_port,
            server_hostname=connect_host if use_tls else None,
            tls=use_tls,
            timeout=connect_timeout,
        )

        conn = Conn(stream)

        if proxy_url is None:
            return conn

        assert proxy_context is not None
        if is_secure:
            target = f"{host}:{port}".encode("ascii")
            connect_headers = [("Host", get_authority_value(host, str(port)))]

            if proxy_url.username or proxy_url.password:
                username = proxy_url.username or ""
                password = proxy_url.password or ""
                credentials = f"{username}:{password}".encode("utf-8")
                auth_value = base64.b64encode(credentials).decode("ascii")
                connect_headers.append(("Proxy-Authorization", f"Basic {auth_value}"))

            self._send_request_headers(
                conn,
                "CONNECT",
                target,
                connect_headers,
            )

            status, _ = self._receive_response_headers(conn, read_timeout=connect_timeout)

            if status < 200 or status > 299:
                conn.close()
                raise ConnectionError(f"Proxy CONNECT failed with status {status}")

            conn.h11 = h11.Connection(h11.CLIENT)

            server_hostname = proxy_context.get("server_hostname") or host
            conn.stream.start_tls(server_hostname=server_hostname)

        return conn

    @staticmethod
    def _prepare_body(
        request: Request,
    ) -> bytes | ABCIterator[bytes] | None:
        if request.body is None:
            return None
        if isinstance(request.body, (bytes, ABCIterator)):
            return request.body
        raise NotImplementedError("Async request bodies are not supported by StdNetworkHandler")

    @staticmethod
    def _response_has_no_body(
        method: str,
        status: int,
        headers: list[tuple[str, str]],
    ) -> bool:
        del headers  # framing is handled by h11; this only covers semantic no-body cases
        if method.upper() == "HEAD":
            return True
        if 100 <= status < 200:
            return True
        if status in (204, 304):
            return True
        return False

    def _write_all(
        self,
        conn: Conn,
        data: bytes,
        *,
        write_timeout: float | None = None,
    ) -> None:
        with map_write_exceptions():
            conn.stream.write_all(data, timeout=write_timeout)

    def _read(
        self,
        conn: Conn,
        n: int,
        *,
        read_timeout: float | None = None,
    ) -> bytes:
        with map_read_exceptions():
            return conn.stream.read(n, timeout=read_timeout)

    def _send_request_headers(
        self,
        conn: Conn,
        method: str,
        target: bytes,
        headers: list[tuple[str, str]],
        *,
        write_timeout: float | None = None,
    ) -> None:
        event = h11.Request(
            method=method.encode("ascii"),
            target=target,
            headers=[
                (
                    k.encode("ascii"),
                    v.encode("latin-1"),
                )
                for k, v in headers
            ],
        )
        self._write_all(
            conn,
            conn.h11.send(event),
            write_timeout=write_timeout,
        )

    def _send_request_body(
        self,
        conn: Conn,
        body: bytes | ABCIterator[bytes] | None,
        *,
        write_timeout: float | None = None,
    ) -> None:
        if isinstance(body, bytes):
            if body:
                self._write_all(
                    conn,
                    conn.h11.send(h11.Data(data=body)),
                    write_timeout=write_timeout,
                )
        elif isinstance(body, ABCIterator):
            for chunk in body:
                if chunk:
                    self._write_all(
                        conn,
                        conn.h11.send(h11.Data(data=chunk)),
                        write_timeout=write_timeout,
                    )

        self._write_all(
            conn,
            conn.h11.send(h11.EndOfMessage()),
            write_timeout=write_timeout,
        )

    def _receive_response_headers(
        self,
        conn: Conn,
        *,
        read_timeout: float | None = None,
    ) -> tuple[int, list[tuple[str, str]]]:
        while True:
            event = conn.h11.next_event()

            if event is h11.NEED_DATA:
                data = self._read(
                    conn,
                    DEFAULT_READ_SIZE,
                    read_timeout=read_timeout,
                )
                if not data:
                    raise ConnectionError("Connection closed while reading response headers")
                conn.h11.receive_data(data)
                continue

            if isinstance(event, h11.InformationalResponse):
                if event.status_code == 101:
                    return event.status_code, [(k.decode("ascii"), v.decode("latin-1")) for k, v in event.headers]

            if isinstance(event, h11.Response):
                status = event.status_code
                resp_headers = [
                    (
                        k.decode("ascii"),
                        v.decode("latin-1"),
                    )
                    for k, v in event.headers
                ]
                return status, resp_headers

            raise ConnectionError(f"Unexpected HTTP event while reading headers: {event!r}")

    def _resolve_timeouts(
        self,
        request: Request,
    ) -> tuple[
        float | None,
        float | None,
        float | None,
        float | None,
    ]:
        """Return (total_timeout, connect_timeout, read_timeout, write_timeout).

        Per-request values from request.context["timeouts"] take priority over
        handler defaults.
        """
        timeouts_context = request.context.get("timeouts", {})

        req_total = timeouts_context.get("total")
        req_connect = timeouts_context.get("connect")
        req_read = timeouts_context.get("read")
        req_write = timeouts_context.get("write")

        total_timeout = req_total if req_total is not None else self.total_timeout
        connect_timeout = req_connect if req_connect is not None else self.connect_timeout
        read_timeout = req_read if req_read is not None else self.read_timeout
        write_timeout = req_write if req_write is not None else self.write_timeout

        return (
            total_timeout,
            connect_timeout,
            read_timeout,
            write_timeout,
        )

    def _acquire_conn_for_request(
        self,
        request: Request,
        key: PoolKey,
        scheme: str,
        host: str,
        port: int,
        *,
        connect_timeout: float | None = None,
    ) -> tuple[Conn, bool]:
        conn = cast(Conn | None, self._pool.acquire(key))
        if conn is not None:
            return conn, True

        try:
            return (
                self._new_conn(
                    request,
                    scheme,
                    host,
                    port,
                    connect_timeout=connect_timeout,
                ),
                False,
            )
        except BaseException:
            self._pool.release_reservation(key)
            raise

    def handle(self, request: Request) -> Response:
        total_timeout, connect_timeout, read_timeout, write_timeout = self._resolve_timeouts(request)
        deadline = None if total_timeout is None else (time.monotonic() + total_timeout)

        scheme = request.url.protocol[:-1]
        host = request.url.hostname
        port = int(request.url.port) if request.url.port != "" else (443 if scheme == "https" else 80)

        key, use_full_url = get_pool_key(request, scheme, host, port)

        def phase_timeout(value: float | None) -> float | None:
            return _min_timeout(value, self._remaining_timeout(deadline))

        conn, from_pool = self._acquire_conn_for_request(
            request,
            key,
            scheme,
            host,
            port,
            connect_timeout=phase_timeout(connect_timeout),
        )

        if use_full_url:
            target = str(request.url).encode("ascii")
        else:
            target = _encode_target(request.url.pathname, request.url.search[1:])
        body = self._prepare_body(request)
        headers = list(request.headers.list())

        proxy_context = request.context.get("network", {}).get("proxy")
        if use_full_url and proxy_context is not None:
            proxy_url_value = proxy_context.get("url")
            if proxy_url_value is not None:
                proxy_url = URL(proxy_url_value) if isinstance(proxy_url_value, str) else proxy_url_value
                if proxy_url.username or proxy_url.password:
                    import base64

                    username = proxy_url.username or ""
                    password = proxy_url.password or ""
                    credentials = f"{username}:{password}".encode("utf-8")
                    auth_value = base64.b64encode(credentials).decode("ascii")
                    headers.append(("Proxy-Authorization", f"Basic {auth_value}"))

        request_wants_close = _header_has_token(headers, "connection", "close")

        try:
            # When using a pooled connection, we might noticed that it was closed by the server when it was idle.
            # In that case, we need to create a new connection and retry the request once.
            try:
                self._send_request_headers(
                    conn,
                    request.method,
                    target,
                    headers,
                    write_timeout=phase_timeout(write_timeout),
                )
            except ConnectionError:
                if not from_pool:
                    raise

                conn.close()
                conn = self._new_conn(
                    request,
                    scheme,
                    host,
                    port,
                    connect_timeout=phase_timeout(connect_timeout),
                )
                self._send_request_headers(
                    conn,
                    request.method,
                    target,
                    headers,
                    write_timeout=phase_timeout(write_timeout),
                )

            self._send_request_body(
                conn,
                body,
                write_timeout=phase_timeout(write_timeout),
            )

            status, resp_headers = self._receive_response_headers(
                conn,
                read_timeout=phase_timeout(read_timeout),
            )

            if status == 101:
                # We won't be able to reuse this connection, so we can release the pool reservation now.
                self._pool.release_reservation(key)
                return Response(
                    status=status,
                    headers=resp_headers,
                    content=None,
                    context={"handoff": ResponseHandoffContext(network_stream=conn.stream)},
                )

        except BaseException:
            self._pool.release(key, conn, reuse=False)
            raise

        no_body_response = self._response_has_no_body(
            request.method,
            status,
            resp_headers,
        )
        response_wants_close = _header_has_token(
            resp_headers,
            "connection",
            "close",
        )

        return Response(
            status=status,
            headers=resp_headers,
            content=StdStream(
                conn,
                self._pool,
                key,
                read_timeout=read_timeout,
                no_body_response=no_body_response,
                must_close=(request_wants_close or response_wants_close),
            ),
        )

    def close(self) -> None:
        self._pool.close_all()



from __future__ import annotations

import asyncio
import contextlib
import socket
from collections.abc import Iterator

import h11

from .._errors import (
    ConnectionError,
    ConnectTimeoutError,
    ReadTimeoutError,
    WriteTimeoutError,
)


@contextlib.contextmanager
def map_connect_exceptions() -> Iterator[None]:
    try:
        yield
    except (asyncio.TimeoutError, socket.timeout) as e:
        raise ConnectTimeoutError("Connection timed out") from e
    except OSError as e:
        if e.errno in (60, 110):
            raise ConnectTimeoutError("Connection timed out") from e
        raise ConnectionError(f"Connection failed: {e}") from e


@contextlib.contextmanager
def map_write_exceptions() -> Iterator[None]:
    try:
        yield
    except (asyncio.TimeoutError, socket.timeout) as e:
        raise WriteTimeoutError("Write operation timed out") from e
    except OSError as e:
        raise ConnectionError(f"Write failed: {e}") from e


@contextlib.contextmanager
def map_read_exceptions() -> Iterator[None]:
    try:
        yield
    except (asyncio.TimeoutError, socket.timeout) as e:
        raise ReadTimeoutError("Read operation timed out") from e
    except (h11.RemoteProtocolError, h11.LocalProtocolError) as e:
        raise ConnectionError(f"HTTP protocol error: {e}") from e
    except OSError as e:
        raise ConnectionError(f"Read failed: {e}") from e



import time

import pytest

from zapros import (
    URL,
    Request,
    Response,
)
from zapros._handlers._mock import (
    Mock as ZaprosMock,
    MockMiddleware,
    MockRouter,
)
from zapros._handlers._retries import (
    DEFAULT_RETRY_STATUS_CODES,
    SAFE_RETRY_METHODS,
    RetryMiddleware,
    RetryPolicy,
)
from zapros._models import Headers


def test_retry_on_503():
    router = MockRouter()
    ZaprosMock().respond(Response(status=503)).once().mount(router)
    ZaprosMock().respond(Response(status=200)).once().mount(router)

    handler = MockMiddleware(router)
    retry_handler = RetryMiddleware(
        handler,
        max_attempts=2,
        backoff_factor=0.0,
    )

    request = Request(
        URL("https://example.com/"),
        "GET",
    )
    response = retry_handler.handle(request)

    assert response.status == 200
    router.verify()


def test_retry_on_network_exception():
    router = MockRouter()
    ZaprosMock().callback(ConnectionError("Network error")).once().mount(router)
    ZaprosMock().respond(Response(status=200)).once().mount(router)

    handler = MockMiddleware(router)
    retry_handler = RetryMiddleware(
        handler,
        max_attempts=2,
        backoff_factor=0.0,
    )

    request = Request(
        URL("https://example.com/"),
        "GET",
    )
    response = retry_handler.handle(request)

    assert response.status == 200
    router.verify()


def test_multiple_retries_until_success():
    router = MockRouter()
    ZaprosMock().respond(Response(status=500)).once().mount(router)
    ZaprosMock().respond(Response(status=500)).once().mount(router)
    ZaprosMock().respond(Response(status=200)).once().mount(router)

    handler = MockMiddleware(router)
    retry_handler = RetryMiddleware(
        handler,
        max_attempts=4,
        backoff_factor=0.0,
    )

    request = Request(
        URL("https://example.com/"),
        "GET",
    )
    response = retry_handler.handle(request)

    assert response.status == 200
    router.verify()


def test_retry_exhaustion_returns_last_response():
    router = MockRouter()
    mock = ZaprosMock().respond(Response(status=503)).expect(3).mount(router)

    handler = MockMiddleware(router)
    retry_handler = RetryMiddleware(
        handler,
        max_attempts=3,
        backoff_factor=0.0,
    )

    request = Request(
        URL("https://example.com/"),
        "GET",
    )
    response = retry_handler.handle(request)

    assert response.status == 503
    mock.verify()


def test_retry_exhaustion_raises_last_exception():
    router = MockRouter()
    mock = ZaprosMock().callback(ConnectionError("Network error")).expect(3).mount(router)

    mock_handler = MockMiddleware(router)
    retry_handler = RetryMiddleware(
        mock_handler,
        max_attempts=3,
        backoff_factor=0.0,
    )

    request = Request(
        URL("https://example.com/"),
        "GET",
    )

    with pytest.raises(ConnectionError):
        retry_handler.handle(request)

    mock.verify()


def test_get_method_retries_on_500():
    router = MockRouter()
    ZaprosMock().respond(Response(status=500)).once().mount(router)
    ZaprosMock().respond(Response(status=200)).once().mount(router)

    handler = MockMiddleware(router)
    retry_handler = RetryMiddleware(
        handler,
        max_attempts=2,
        backoff_factor=0.0,
    )

    request = Request(
        URL("https://example.com/"),
        "GET",
    )
    response = retry_handler.handle(request)

    assert response.status == 200
    router.verify()


def test_post_method_does_not_retry():
    router = MockRouter()
    mock = ZaprosMock().respond(Response(status=500)).once().mount(router)

    handler = MockMiddleware(router)
    retry_handler = RetryMiddleware(
        handler,
        max_attempts=3,
        backoff_factor=0.0,
    )

    request = Request(
        URL("https://example.com/"),
        "POST",
        text="body",
    )
    response = retry_handler.handle(request)

    assert response.status == 500
    mock.verify()


def test_put_method_retries():
    router = MockRouter()
    ZaprosMock().respond(Response(status=500)).once().mount(router)
    ZaprosMock().respond(Response(status=200)).once().mount(router)

    handler = MockMiddleware(router)
    retry_handler = RetryMiddleware(
        handler,
        max_attempts=2,
        backoff_factor=0.0,
    )

    request = Request(
        URL("https://example.com/"),
        "PUT",
        text="body",
    )
    response = retry_handler.handle(request)

    assert response.status == 200
    router.verify()


def test_delete_method_retries():
    router = MockRouter()
    ZaprosMock().respond(Response(status=500)).once().mount(router)
    ZaprosMock().respond(Response(status=200)).once().mount(router)

    handler = MockMiddleware(router)
    retry_handler = RetryMiddleware(
        handler,
        max_attempts=2,
        backoff_factor=0.0,
    )

    request = Request(
        URL("https://example.com/"),
        "DELETE",
    )
    response = retry_handler.handle(request)

    assert response.status == 200
    router.verify()


def test_retry_on_429_rate_limit():
    router = MockRouter()
    ZaprosMock().respond(Response(status=429)).once().mount(router)
    ZaprosMock().respond(Response(status=200)).once().mount(router)

    handler = MockMiddleware(router)
    retry_handler = RetryMiddleware(
        handler,
        max_attempts=2,
        backoff_factor=0.0,
    )

    request = Request(
        URL("https://example.com/"),
        "GET",
    )
    response = retry_handler.handle(request)

    assert response.status == 200
    router.verify()


def test_retry_on_502_bad_gateway():
    router = MockRouter()
    ZaprosMock().respond(Response(status=502)).once().mount(router)
    ZaprosMock().respond(Response(status=200)).once().mount(router)

    handler = MockMiddleware(router)
    retry_handler = RetryMiddleware(
        handler,
        max_attempts=2,
        backoff_factor=0.0,
    )

    request = Request(
        URL("https://example.com/"),
        "GET",
    )
    response = retry_handler.handle(request)

    assert response.status == 200
    router.verify()


def test_retry_on_504_gateway_timeout():
    router = MockRouter()
    ZaprosMock().respond(Response(status=504)).once().mount(router)
    ZaprosMock().respond(Response(status=200)).once().mount(router)

    handler = MockMiddleware(router)
    retry_handler = RetryMiddleware(
        handler,
        max_attempts=2,
        backoff_factor=0.0,
    )

    request = Request(
        URL("https://example.com/"),
        "GET",
    )
    response = retry_handler.handle(request)

    assert response.status == 200
    router.verify()


def test_no_retry_on_404():
    router = MockRouter()
    mock = ZaprosMock().respond(Response(status=404)).once().mount(router)

    handler = MockMiddleware(router)
    retry_handler = RetryMiddleware(
        handler,
        max_attempts=3,
        backoff_factor=0.0,
    )

    request = Request(
        URL("https://example.com/"),
        "GET",
    )
    response = retry_handler.handle(request)

    assert response.status == 404
    mock.verify()


def test_no_retry_on_200():
    router = MockRouter()
    mock = ZaprosMock().respond(Response(status=200)).once().mount(router)

    handler = MockMiddleware(router)
    retry_handler = RetryMiddleware(
        handler,
        max_attempts=3,
        backoff_factor=0.0,
    )

    request = Request(
        URL("https://example.com/"),
        "GET",
    )
    response = retry_handler.handle(request)

    assert response.status == 200
    mock.verify()


def test_no_retry_on_301():
    router = MockRouter()
    mock = (
        ZaprosMock()
        .respond(
            Response(
                status=301,
                headers={"Location": "/other"},
            )
        )
        .once()
        .mount(router)
    )

    handler = MockMiddleware(router)
    retry_handler = RetryMiddleware(
        handler,
        max_attempts=3,
        backoff_factor=0.0,
    )

    request = Request(
        URL("https://example.com/"),
        "GET",
    )
    response = retry_handler.handle(request)

    assert response.status == 301
    mock.verify()


def test_max_attempts_one_means_no_retry():
    router = MockRouter()
    mock = ZaprosMock().respond(Response(status=500)).once().mount(router)

    handler = MockMiddleware(router)
    retry_handler = RetryMiddleware(
        handler,
        max_attempts=1,
        backoff_factor=0.0,
    )

    request = Request(
        URL("https://example.com/"),
        "GET",
    )
    response = retry_handler.handle(request)

    assert response.status == 500
    mock.verify()


def test_exponential_backoff_timing():
    timestamps = []

    def handle_with_timestamps(
        request: Request,
    ) -> Response:
        timestamps.append(time.time())
        if len(timestamps) < 3:
            return Response(
                status=500,
                headers=Headers({}),
                content=None,
            )
        return Response(
            status=200,
            headers=Headers({}),
            content=None,
        )

    router = MockRouter()
    ZaprosMock().callback(handle_with_timestamps).mount(router)
    handler = MockMiddleware(router)
    retry_handler = RetryMiddleware(
        handler,
        max_attempts=3,
        backoff_factor=0.1,
        backoff_jitter=0.0,
    )

    request = Request(
        URL("https://example.com/"),
        "GET",
    )
    response = retry_handler.handle(request)

    assert response.status == 200
    assert len(timestamps) == 3

    delay1 = timestamps[1] - timestamps[0]
    delay2 = timestamps[2] - timestamps[1]

    assert 0.08 <= delay1 <= 0.15
    assert 0.15 <= delay2 <= 0.25


def test_backoff_max_cap():
    timestamps = []

    def handle_with_timestamps(
        request: Request,
    ) -> Response:
        timestamps.append(time.time())
        if len(timestamps) < 3:
            return Response(
                status=500,
                headers=Headers({}),
                content=None,
            )
        return Response(
            status=200,
            headers=Headers({}),
            content=None,
        )

    router = MockRouter()
    ZaprosMock().callback(handle_with_timestamps).mount(router)
    handler = MockMiddleware(router)
    retry_handler = RetryMiddleware(
        handler,
        max_attempts=3,
        backoff_factor=1.0,
        backoff_max=0.2,
        backoff_jitter=0.0,
    )

    request = Request(
        URL("https://example.com/"),
        "GET",
    )
    response = retry_handler.handle(request)

    assert response.status == 200

    delay2 = timestamps[2] - timestamps[1]
    assert delay2 <= 0.25


def test_custom_retry_policy():
    class CustomRetryPolicy:
        def should_retry(
            self,
            *,
            request: Request,
            response: Response | None,
            error: Exception | None,
            attempt: int,
        ) -> bool:
            if response and response.status == 418:
                return True
            return False

    router = MockRouter()
    ZaprosMock().respond(Response(status=418)).once().mount(router)
    ZaprosMock().respond(Response(status=200)).once().mount(router)

    handler = MockMiddleware(router)
    retry_handler = RetryMiddleware(
        handler,
        policy=CustomRetryPolicy(),
        max_attempts=2,
        backoff_factor=0.0,
    )

    request = Request(
        URL("https://example.com/"),
        "POST",
        text="body",
    )
    response = retry_handler.handle(request)

    assert response.status == 200
    router.verify()


def test_custom_policy_overrides_defaults():
    class AlwaysRetryPolicy:
        def should_retry(
            self,
            *,
            request: Request,
            response: Response | None,
            error: Exception | None,
            attempt: int,
        ) -> bool:
            return True

    router = MockRouter()
    ZaprosMock().respond(Response(status=404)).once().mount(router)
    ZaprosMock().respond(Response(status=404)).once().mount(router)
    ZaprosMock().respond(Response(status=200)).once().mount(router)

    handler = MockMiddleware(router)
    retry_handler = RetryMiddleware(
        handler,
        policy=AlwaysRetryPolicy(),
        max_attempts=3,
        backoff_factor=0.0,
    )

    request = Request(
        URL("https://example.com/"),
        "GET",
    )
    response = retry_handler.handle(request)

    assert response.status == 200
    router.verify()


def test_second_attempt_succeeds():
    router = MockRouter()
    ZaprosMock().respond(Response(status=500)).once().mount(router)
    ZaprosMock().respond(Response(status=200)).once().mount(router)

    handler = MockMiddleware(router)
    retry_handler = RetryMiddleware(
        handler,
        max_attempts=2,
        backoff_factor=0.0,
    )

    request = Request(
        URL("https://example.com/"),
        "GET",
    )
    response = retry_handler.handle(request)

    assert response.status == 200
    router.verify()


@pytest.mark.asyncio
async def test_async_retry_on_503():
    router = MockRouter()
    ZaprosMock().respond(Response(status=503)).once().mount(router)
    ZaprosMock().respond(Response(status=200)).once().mount(router)

    handler = MockMiddleware(router)
    retry_handler = RetryMiddleware(
        handler,
        max_attempts=2,
        backoff_factor=0.0,
    )

    request = Request(
        URL("https://example.com/"),
        "GET",
    )
    response = await retry_handler.ahandle(request)

    assert response.status == 200
    router.verify()


@pytest.mark.asyncio
async def test_async_retry_on_network_exception():
    router = MockRouter()
    ZaprosMock().callback(ConnectionError("Network error")).once().mount(router)
    ZaprosMock().respond(Response(status=200)).once().mount(router)

    handler = MockMiddleware(router)
    retry_handler = RetryMiddleware(
        handler,
        max_attempts=2,
        backoff_factor=0.0,
    )

    request = Request(
        URL("https://example.com/"),
        "GET",
    )
    response = await retry_handler.ahandle(request)

    assert response.status == 200
    router.verify()


@pytest.mark.asyncio
async def test_async_post_does_not_retry():
    router = MockRouter()
    mock = ZaprosMock().respond(Response(status=500)).once().mount(router)

    handler = MockMiddleware(router)
    retry_handler = RetryMiddleware(
        handler,
        max_attempts=3,
        backoff_factor=0.0,
    )

    request = Request(
        URL("https://example.com/"),
        "POST",
        text="body",
    )
    response = await retry_handler.ahandle(request)

    assert response.status == 500
    mock.verify()


@pytest.mark.asyncio
async def test_async_exponential_backoff():
    timestamps = []

    def handle_with_timestamps(
        request: Request,
    ) -> Response:
        timestamps.append(time.time())
        if len(timestamps) < 3:
            return Response(
                status=500,
                headers=Headers({}),
                content=None,
            )
        return Response(
            status=200,
            headers=Headers({}),
            content=None,
        )

    router = MockRouter()
    ZaprosMock().callback(handle_with_timestamps).mount(router)
    handler = MockMiddleware(router)
    retry_handler = RetryMiddleware(
        handler,
        max_attempts=3,
        backoff_factor=0.1,
        backoff_jitter=0.0,
    )

    request = Request(
        URL("https://example.com/"),
        "GET",
    )
    response = await retry_handler.ahandle(request)

    assert response.status == 200
    assert len(timestamps) == 3

    delay1 = timestamps[1] - timestamps[0]
    delay2 = timestamps[2] - timestamps[1]

    assert 0.08 <= delay1 <= 0.15
    assert 0.15 <= delay2 <= 0.25


def test_retry_protocol_check():
    class CustomPolicy:
        def should_retry(
            self,
            *,
            request: Request,
            response: Response | None,
            error: Exception | None,
            attempt: int,
        ) -> bool:
            return False

    policy = CustomPolicy()
    assert isinstance(policy, RetryPolicy)


def test_default_constants():
    assert 429 in DEFAULT_RETRY_STATUS_CODES
    assert 500 in DEFAULT_RETRY_STATUS_CODES
    assert 502 in DEFAULT_RETRY_STATUS_CODES
    assert 503 in DEFAULT_RETRY_STATUS_CODES
    assert 504 in DEFAULT_RETRY_STATUS_CODES

    assert "GET" in SAFE_RETRY_METHODS
    assert "HEAD" in SAFE_RETRY_METHODS
    assert "PUT" in SAFE_RETRY_METHODS
    assert "DELETE" in SAFE_RETRY_METHODS
    assert "OPTIONS" in SAFE_RETRY_METHODS
    assert "TRACE" in SAFE_RETRY_METHODS
    assert "POST" not in SAFE_RETRY_METHODS
    assert "PATCH" not in SAFE_RETRY_METHODS


def test_head_method_retries():
    router = MockRouter()
    ZaprosMock().respond(Response(status=500)).once().mount(router)
    ZaprosMock().respond(Response(status=200)).once().mount(router)

    handler = MockMiddleware(router)
    retry_handler = RetryMiddleware(
        handler,
        max_attempts=2,
        backoff_factor=0.0,
    )

    request = Request(
        URL("https://example.com/"),
        "HEAD",
    )
    response = retry_handler.handle(request)

    assert response.status == 200
    router.verify()


def test_options_method_retries():
    router = MockRouter()
    ZaprosMock().respond(Response(status=500)).once().mount(router)
    ZaprosMock().respond(Response(status=200)).once().mount(router)

    handler = MockMiddleware(router)
    retry_handler = RetryMiddleware(
        handler,
        max_attempts=2,
        backoff_factor=0.0,
    )

    request = Request(
        URL("https://example.com/"),
        "OPTIONS",
    )
    response = retry_handler.handle(request)

    assert response.status == 200
    router.verify()


def test_trace_method_retries():
    router = MockRouter()
    ZaprosMock().respond(Response(status=500)).once().mount(router)
    ZaprosMock().respond(Response(status=200)).once().mount(router)

    handler = MockMiddleware(router)
    retry_handler = RetryMiddleware(
        handler,
        max_attempts=2,
        backoff_factor=0.0,
    )

    request = Request(
        URL("https://example.com/"),
        "TRACE",
    )
    response = retry_handler.handle(request)

    assert response.status == 200
    router.verify()


def test_patch_method_does_not_retry():
    router = MockRouter()
    mock = ZaprosMock().respond(Response(status=500)).once().mount(router)

    handler = MockMiddleware(router)
    retry_handler = RetryMiddleware(
        handler,
        max_attempts=3,
        backoff_factor=0.0,
    )

    request = Request(
        URL("https://example.com/"),
        "PATCH",
        text="body",
    )
    response = retry_handler.handle(request)

    assert response.status == 500
    mock.verify()


def test_post_retries_on_connection_error_before_transmission():
    router = MockRouter()
    ZaprosMock().callback(ConnectionError("Connection refused")).once().mount(router)
    ZaprosMock().respond(Response(status=200)).once().mount(router)

    handler = MockMiddleware(router)
    retry_handler = RetryMiddleware(
        handler,
        max_attempts=2,
        backoff_factor=0.0,
    )

    request = Request(
        URL("https://example.com/"),
        "POST",
        text="body",
    )
    response = retry_handler.handle(request)

    assert response.status == 200
    router.verify()


def test_post_retries_on_connection_refused_error():
    router = MockRouter()
    ZaprosMock().callback(ConnectionRefusedError("Connection refused")).once().mount(router)
    ZaprosMock().respond(Response(status=200)).once().mount(router)

    handler = MockMiddleware(router)
    retry_handler = RetryMiddleware(
        handler,
        max_attempts=2,
        backoff_factor=0.0,
    )

    request = Request(
        URL("https://example.com/"),
        "POST",
        text="body",
    )
    response = retry_handler.handle(request)

    assert response.status == 200
    router.verify()


def test_post_does_not_retry_on_read_timeout():
    class ReadTimeoutError(Exception):
        pass

    router = MockRouter()
    mock = ZaprosMock().callback(ReadTimeoutError("Read timeout")).once().mount(router)

    mock_handler = MockMiddleware(router)
    retry_handler = RetryMiddleware(
        mock_handler,
        max_attempts=3,
        backoff_factor=0.0,
    )

    request = Request(
        URL("https://example.com/"),
        "POST",
        text="body",
    )

    with pytest.raises(ReadTimeoutError):
        retry_handler.handle(request)

    mock.verify()



from __future__ import annotations

import threading
import time
from threading import Semaphore
from collections import deque
from dataclasses import dataclass

from ._base_pool import IdlePoolConnection, PoolConnection, PoolKey


@dataclass
class _HostState:
    semaphore: Semaphore
    refs: int = 0
    idle: deque[IdlePoolConnection] | None = None


class ConnPool:
    def __init__(
        self,
        *,
        max_connections_per_host: int = 10,
        max_idle_per_host: int = 10,
        max_idle_seconds: float = 30.0,
    ) -> None:
        self._max_connections_per_host = max_connections_per_host
        self._max_idle_per_host = max_idle_per_host
        self._max_age = max_idle_seconds

        self._lock = threading.Lock()
        self._states: dict[PoolKey, _HostState] = {}
        self._closed = False

    def _get_state(self, key: PoolKey) -> _HostState:
        with self._lock:
            state = self._states.get(key)
            if state is None:
                state = _HostState(semaphore=Semaphore(self._max_connections_per_host))
                self._states[key] = state

            state.refs += 1
            return state

    def _release_state_ref(self, key: PoolKey, state: _HostState) -> None:
        with self._lock:
            if self._states.get(key) is not state:
                return

            if state.refs <= 1:
                state.refs = 0
            else:
                state.refs -= 1

            if state.refs == 0 and state.idle is None and state.semaphore._value >= self._max_connections_per_host:
                self._states.pop(key, None)

    def _take_idle_connection(
        self,
        key: PoolKey,
        now: float,
    ) -> tuple[PoolConnection | None, list[PoolConnection]]:
        stale: list[PoolConnection] = []

        with self._lock:
            if self._closed:
                raise RuntimeError("Pool is closed")

            state = self._states.get(key)
            dq = None if state is None else state.idle

            if not dq:
                return None, stale

            while dq:
                item = dq.pop()
                if (now - item.last_used) <= self._max_age and item.conn.can_reuse():
                    if not dq:
                        state.idle = None  # type: ignore[reportOptionalMemberAccess]
                    return item.conn, stale

                stale.append(item.conn)

            state.idle = None  # type: ignore[reportOptionalMemberAccess]
            return None, stale

    def _store_idle_connection(
        self,
        key: PoolKey,
        conn: PoolConnection,
    ) -> tuple[bool, list[PoolConnection]]:
        """
        Adds conn to the idle deque for key.

        Returns:
            kept:
                True when conn remained in the idle pool.
                False when it was rejected/evicted.
            to_close:
                Connections removed from the idle pool and needing close().
        """
        with self._lock:
            if self._closed:
                return False, [conn]

            state = self._states.get(key)
            if state is None:
                state = _HostState(semaphore=Semaphore(self._max_connections_per_host))
                self._states[key] = state

            dq = state.idle
            if dq is None:
                dq = state.idle = deque()

            dq.append(IdlePoolConnection(conn=conn))

            to_close: list[PoolConnection] = []
            while len(dq) > self._max_idle_per_host:
                to_close.append(dq.popleft().conn)

            return conn not in to_close, to_close

    def acquire(
        self,
        key: PoolKey,
    ) -> PoolConnection | None:
        """
        Acquires a connection from the pool for the given key.

        Tries to get a slot for the key, then returns an idle connection if available
        and valid, otherwise returns None.

        When None is returned the caller is expected to create a new connection and
        call release() with reuse=False when done.
        """
        state = self._get_state(key)
        sem = state.semaphore
        sem.acquire()
        release_sem = True

        try:
            conn, to_close = self._take_idle_connection(
                key,
                time.monotonic(),
            )
            _close_many_quietly(to_close)

            release_sem = False
            return conn

        except BaseException:
            if release_sem:
                sem.release()
            raise

        finally:
            self._release_state_ref(key, state)

    def release_reservation(self, key: PoolKey) -> None:
        state = self._get_state(key)
        try:
            state.semaphore.release()
        finally:
            self._release_state_ref(key, state)

    def release(
        self,
        key: PoolKey,
        conn: PoolConnection,
        *,
        reuse: bool,
    ) -> None:
        """
        Releases a connection back to the pool for the given key.

        If reuse is False, the connection will be closed and not added back to the pool.
        """
        state = self._get_state(key)

        try:
            if not reuse:
                _close_quietly(conn)
                return

            _, to_close = self._store_idle_connection(key, conn)
            _close_many_quietly(to_close)

        finally:
            state.semaphore.release()
            self._release_state_ref(key, state)

    def close_all(self) -> None:
        """
        Closes all idle connections in the pool and prevents new connections
        from being added.
        """
        with self._lock:
            if self._closed:
                return

            self._closed = True
            states, self._states = self._states, {}

        for state in states.values():
            for item in state.idle or ():
                _close_quietly(item.conn)


def _close_many_quietly(conns: list[PoolConnection]) -> None:
    for conn in conns:
        _close_quietly(conn)


def _close_quietly(conn: PoolConnection) -> None:
    try:
        conn.close()
    except Exception:
        pass



from __future__ import annotations

import asyncio
import contextlib
from collections.abc import (
    AsyncGenerator,
    AsyncIterator,
    Generator,
    Iterator as ABCIterator,
)
from dataclasses import dataclass
from typing import Any

from typing_extensions import override

from zapros._constants import (
    DEFAULT_PORTS,
)

from .._models import (
    AsyncClosableStream,
    Request,
    Response,
)
from ._async_base import (
    AsyncBaseHandler,
)


@dataclass(slots=True)
class _QueuedBodyChunk:
    body: bytes
    more_body: bool


@dataclass(slots=True)
class _LifespanSession:
    state: dict[str, Any]
    receive_queue: "asyncio.Queue[dict[str, Any]]"
    send_queue: "asyncio.Queue[dict[str, Any]]"
    task: "asyncio.Task[None]"


class _RequestBodySource:
    def __init__(
        self,
        source: bytes | ABCIterator[bytes] | AsyncIterator[bytes] | None,
    ) -> None:
        self._source = source
        self._done = False
        self._closed = False

    @property
    def done(self) -> bool:
        return self._done

    async def next_event(
        self,
    ) -> dict[str, Any]:
        if self._done:
            raise RuntimeError("Request body source is already exhausted")

        source = self._source

        if source is None:
            self._done = True
            return {
                "type": "http.request",
                "body": b"",
                "more_body": False,
            }

        if isinstance(source, bytes):
            self._done = True
            self._source = None
            return {
                "type": "http.request",
                "body": source,
                "more_body": False,
            }

        if isinstance(source, ABCIterator):
            raise RuntimeError("Synchronous iterators are not supported as request body sources by AsgiHandler")

        async_iter = source
        try:
            chunk = await anext(async_iter)
        except StopAsyncIteration:
            self._done = True
            self._source = None
            return {
                "type": "http.request",
                "body": b"",
                "more_body": False,
            }

        return {
            "type": "http.request",
            "body": bytes(chunk),
            "more_body": True,
        }

    async def aclose(self) -> None:
        if self._closed:
            return

        self._closed = True
        source = self._source
        self._source = None
        self._done = True

        if source is None or isinstance(source, bytes):
            return

        if isinstance(source, AsyncGenerator):
            await source.aclose()
            return

        if isinstance(source, Generator):
            source.close()


class _ResponseStream(AsyncClosableStream):
    def __init__(
        self,
        body_queue: "asyncio.Queue[_QueuedBodyChunk | BaseException | None]",
        app_task: "asyncio.Task[None]",
        disconnect_event: "asyncio.Event",
    ) -> None:
        self._body_queue = body_queue
        self._app_task = app_task
        self._disconnect_event = disconnect_event
        self._closed = False

    def __aiter__(
        self,
    ) -> AsyncIterator[bytes]:
        return self

    async def _finalize(self, *, abort: bool) -> BaseException | None:
        if self._closed:
            return None

        self._closed = True
        self._disconnect_event.set()

        if abort:
            if not self._app_task.done():
                self._app_task.cancel()

            with contextlib.suppress(asyncio.CancelledError):
                await self._app_task
            return None

        try:
            await self._app_task
        except BaseException as exc:
            return exc

        return None

    async def __anext__(self) -> bytes:
        while True:
            if self._closed:
                raise StopAsyncIteration

            item = await self._body_queue.get()

            if item is None:
                task_error = await self._finalize(abort=False)
                if task_error is not None:
                    raise task_error
                raise StopAsyncIteration

            if isinstance(item, BaseException):
                task_error = await self._finalize(abort=False)
                raise task_error or item

            if item.body:
                return item.body

            if not item.more_body:
                task_error = await self._finalize(abort=False)
                if task_error is not None:
                    raise task_error
                raise StopAsyncIteration

    @override
    async def aclose(self) -> None:
        await self._finalize(abort=True)


class AsgiHandler(AsyncBaseHandler):
    def __init__(
        self,
        app: Any,
        *,
        client: tuple[str, int] | None = ("127.0.0.1", 123),
        root_path: str = "",
        http_version: str = "1.1",
        enable_lifespan: bool = True,
        startup_timeout: float | None = 10.0,
        shutdown_timeout: float | None = 10.0,
    ) -> None:
        self.app = app
        self.client = client
        self.root_path = root_path
        self.http_version = http_version
        self.enable_lifespan = enable_lifespan
        self.startup_timeout = startup_timeout
        self.shutdown_timeout = shutdown_timeout

        self._lifespan_lock = asyncio.Lock()
        self._lifespan_started = False
        self._lifespan_supported = True
        self._lifespan_state: dict[str, Any] = {}
        self._lifespan_session: _LifespanSession | None = None

    @staticmethod
    def _iter_request_body(
        request: Request,
    ) -> bytes | ABCIterator[bytes] | AsyncIterator[bytes] | None:
        if request.body is None:
            return None
        if isinstance(request.body, bytes):
            return request.body
        if isinstance(request.body, ABCIterator):
            return request.body
        return request.body

    @staticmethod
    def _get_raw_path_bytes(
        request: Request,
    ) -> bytes | None:
        raw_path = getattr(request, "raw_path", None)
        if raw_path is None:
            raw_path = getattr(
                request.url,
                "raw_path",
                None,
            )

        if isinstance(raw_path, memoryview):
            return raw_path.tobytes()
        if isinstance(raw_path, (bytes, bytearray)):
            return bytes(raw_path)

        return None

    @staticmethod
    def _get_query_string_bytes(
        request: Request,
    ) -> bytes:
        raw_query = getattr(
            request,
            "raw_query_string",
            None,
        )
        if raw_query is None:
            raw_query = getattr(
                request.url,
                "raw_query",
                None,
            )
        if raw_query is None:
            raw_query = request.url.search

        if raw_query in (None, ""):
            return b""

        if isinstance(raw_query, memoryview):
            raw_query_bytes = raw_query.tobytes()
            return raw_query_bytes[1:] if raw_query_bytes.startswith(b"?") else raw_query_bytes
        if isinstance(
            raw_query,
            (bytes, bytearray),
        ):
            raw_query_bytes = bytes(raw_query)
            return raw_query_bytes[1:] if raw_query_bytes.startswith(b"?") else raw_query_bytes

        query_string = str(raw_query)
        if query_string.startswith("?"):
            query_string = query_string[1:]
        return query_string.encode("ascii")

    @staticmethod
    async def _wait_for_lifespan_message(
        session: _LifespanSession,
        *,
        timeout: float | None,
    ) -> dict[str, Any] | None:
        message_task = asyncio.create_task(session.send_queue.get())
        try:
            (
                done,
                _,
            ) = await asyncio.wait(
                {
                    message_task,
                    session.task,
                },
                timeout=timeout,
                return_when=asyncio.FIRST_COMPLETED,
            )
            if not done:
                raise asyncio.TimeoutError

            if message_task in done:
                return message_task.result()

            if not session.send_queue.empty():
                return session.send_queue.get_nowait()

            return None
        finally:
            if not message_task.done():
                message_task.cancel()
                with contextlib.suppress(asyncio.CancelledError):
                    await message_task

    @staticmethod
    async def _cancel_task(
        task: "asyncio.Task[Any]",
    ) -> None:
        if not task.done():
            task.cancel()
        with contextlib.suppress(
            asyncio.CancelledError,
            Exception,
        ):
            await task

    @staticmethod
    async def _await_task(
        task: "asyncio.Task[Any]",
        *,
        timeout: float | None,
        timeout_message: str,
    ) -> None:
        try:
            if timeout is None:
                await task
            else:
                await asyncio.wait_for(
                    asyncio.shield(task),
                    timeout=timeout,
                )
        except asyncio.TimeoutError as exc:
            if not task.done():
                task.cancel()
                with contextlib.suppress(
                    asyncio.CancelledError,
                    Exception,
                ):
                    await task
            raise RuntimeError(timeout_message) from exc

    def _build_scope(self, request: Request) -> dict[str, Any]:
        headers = [
            (
                k.lower().encode("ascii"),
                v.encode("latin-1"),
            )
            for k, v in request.headers.list()
        ]

        raw_path = self._get_raw_path_bytes(request)
        query_string = self._get_query_string_bytes(request)
        server_port = request.url.port or DEFAULT_PORTS.get(
            request.url.protocol[:-1],
            80,
        )

        scope: dict[str, Any] = {
            "type": "http",
            "asgi": {
                "version": "3.0",
                "spec_version": "2.4",
            },
            "http_version": self.http_version,
            "method": request.method.upper(),
            "scheme": request.url.protocol[:-1],
            "path": request.url.pathname or "/",
            "query_string": query_string,
            "root_path": self.root_path,
            "headers": headers,
            "client": self.client,
            "server": (
                request.url.hostname,
                server_port,
            ),
        }

        if raw_path is not None:
            scope["raw_path"] = raw_path

        if self._lifespan_supported:
            scope["state"] = self._lifespan_state.copy()

        return scope

    def _raise_if_lifespan_task_exited(
        self,
    ) -> None:
        session = self._lifespan_session
        if not self._lifespan_started or session is None or not session.task.done():
            return

        self._lifespan_started = False
        self._lifespan_state = {}
        self._lifespan_session = None

        exc = session.task.exception()
        if exc is None:
            raise RuntimeError("ASGI lifespan task exited unexpectedly before shutdown")
        raise RuntimeError("ASGI lifespan task crashed after startup") from exc

    async def _run_lifespan(
        self,
    ) -> None:
        if not self.enable_lifespan or self._lifespan_started or not self._lifespan_supported:
            return

        async with self._lifespan_lock:
            if self._lifespan_started or not self.enable_lifespan or not self._lifespan_supported:
                return

            state: dict[str, Any] = {}
            receive_queue: asyncio.Queue[dict[str, Any]] = asyncio.Queue()
            send_queue: asyncio.Queue[dict[str, Any]] = asyncio.Queue()

            async def receive() -> dict[str, Any]:
                return await receive_queue.get()

            async def send(
                message: dict[str, Any],
            ) -> None:
                await send_queue.put(message)

            scope = {
                "type": "lifespan",
                "asgi": {
                    "version": "3.0",
                    "spec_version": "2.0",
                },
                "state": state,
            }

            task = asyncio.create_task(self.app(scope, receive, send))
            session = _LifespanSession(
                state=state,
                receive_queue=receive_queue,
                send_queue=send_queue,
                task=task,
            )
            self._lifespan_session = session

            await receive_queue.put({"type": "lifespan.startup"})

            try:
                message = await self._wait_for_lifespan_message(
                    session,
                    timeout=self.startup_timeout,
                )
            except asyncio.TimeoutError as exc:
                self._lifespan_session = None
                self._lifespan_state = {}
                await self._cancel_task(task)
                raise RuntimeError("ASGI lifespan startup timed out") from exc

            if message is None:
                self._lifespan_supported = False
                self._lifespan_session = None
                self._lifespan_state = {}
                return

            msg_type = message["type"]

            if msg_type == "lifespan.startup.complete":
                self._lifespan_state = state
                self._lifespan_started = True
                return

            self._lifespan_session = None
            self._lifespan_state = {}
            await self._cancel_task(task)

            if msg_type == "lifespan.startup.failed":
                detail = message.get("message", "")
                raise RuntimeError(f"ASGI lifespan startup failed: {detail}")

            raise RuntimeError(f"Unexpected ASGI lifespan startup message: {msg_type!r}")

    async def ahandle(self, request: Request) -> Response:
        await self._run_lifespan()
        self._raise_if_lifespan_task_exited()

        scope = self._build_scope(request)
        request_body = _RequestBodySource(self._iter_request_body(request))

        response_started = asyncio.Event()
        response_complete = asyncio.Event()
        disconnect_event = asyncio.Event()

        response_status: int | None = None
        response_headers: list[tuple[str, str]] | None = None
        body_queue: asyncio.Queue[_QueuedBodyChunk | BaseException | None] = asyncio.Queue()
        app_error: BaseException | None = None

        receive_lock = asyncio.Lock()
        send_lock = asyncio.Lock()

        async def receive() -> dict[str, Any]:
            async with receive_lock:
                if response_complete.is_set() or disconnect_event.is_set():
                    return {"type": "http.disconnect"}

                if not request_body.done:
                    return await request_body.next_event()

            waiter1 = asyncio.create_task(response_complete.wait())
            waiter2 = asyncio.create_task(disconnect_event.wait())
            try:
                await asyncio.wait(
                    {waiter1, waiter2},
                    return_when=asyncio.FIRST_COMPLETED,
                )
            finally:
                waiter1.cancel()
                waiter2.cancel()
                with contextlib.suppress(asyncio.CancelledError):
                    await waiter1
                with contextlib.suppress(asyncio.CancelledError):
                    await waiter2

            return {"type": "http.disconnect"}

        async def send(
            message: dict[str, Any],
        ) -> None:
            nonlocal response_status, response_headers

            msg_type = message["type"]

            async with send_lock:
                if response_complete.is_set():
                    return

                if disconnect_event.is_set():
                    raise OSError("Connection closed")

                if msg_type == "http.response.start":
                    if response_started.is_set():
                        raise RuntimeError("ASGI application sent multiple http.response.start messages")

                    if bool(
                        message.get(
                            "trailers",
                            False,
                        )
                    ):
                        raise NotImplementedError("ASGI response trailers are not supported by AsgiHandler")

                    response_status = int(message["status"])
                    raw_headers = message.get(
                        "headers",
                        [],
                    )
                    response_headers = [
                        (
                            bytes(k).decode("ascii"),
                            bytes(v).decode("latin-1"),
                        )
                        for k, v in raw_headers
                    ]
                    response_started.set()
                    return

                if msg_type == "http.response.body":
                    if not response_started.is_set():
                        raise RuntimeError("ASGI application sent http.response.body before http.response.start")

                    body = bytes(message.get("body", b""))
                    more_body = bool(
                        message.get(
                            "more_body",
                            False,
                        )
                    )
                    await body_queue.put(
                        _QueuedBodyChunk(
                            body=body,
                            more_body=more_body,
                        )
                    )

                    if not more_body:
                        response_complete.set()
                        await body_queue.put(None)
                    return

                if msg_type == "http.response.trailers":
                    raise NotImplementedError("ASGI response trailers are not supported by AsgiHandler")

                raise RuntimeError(f"Unsupported ASGI send message type: {msg_type!r}")

        async def run_app() -> None:
            nonlocal app_error
            try:
                await self.app(scope, receive, send)

                if not response_started.is_set():
                    raise RuntimeError("ASGI application returned without sending http.response.start")

                if not response_complete.is_set():
                    await send(
                        {
                            "type": "http.response.body",
                            "body": b"",
                            "more_body": False,
                        }
                    )
            except asyncio.CancelledError:
                raise
            except BaseException as exc:
                app_error = exc
                if not response_started.is_set():
                    response_started.set()
                    return

                if not response_complete.is_set():
                    await body_queue.put(exc)
                    response_complete.set()
                    await body_queue.put(None)
                    return

                raise
            finally:
                with contextlib.suppress(Exception):
                    await request_body.aclose()

        app_task = asyncio.create_task(run_app())

        await response_started.wait()

        if response_status is None or response_headers is None:
            with contextlib.suppress(asyncio.CancelledError):
                await app_task
            assert app_error is not None
            raise app_error

        return Response(
            status=response_status,
            headers=response_headers,
            content=_ResponseStream(
                body_queue=body_queue,
                app_task=app_task,
                disconnect_event=disconnect_event,
            ),
        )

    async def aclose(self) -> None:
        if not self.enable_lifespan:
            return
        if not self._lifespan_started or not self._lifespan_supported:
            return

        async with self._lifespan_lock:
            if not self._lifespan_started or not self._lifespan_supported:
                return

            session = self._lifespan_session
            if session is None:
                self._lifespan_started = False
                self._lifespan_state = {}
                return

            await session.receive_queue.put({"type": "lifespan.shutdown"})

            try:
                message = await self._wait_for_lifespan_message(
                    session,
                    timeout=self.shutdown_timeout,
                )
            except asyncio.TimeoutError as exc:
                self._lifespan_started = False
                self._lifespan_state = {}
                self._lifespan_session = None
                await self._cancel_task(session.task)
                raise RuntimeError("ASGI lifespan shutdown timed out") from exc

            self._lifespan_started = False
            self._lifespan_state = {}
            self._lifespan_session = None

            if message is None:
                exc = session.task.exception()
                if exc is None:
                    raise RuntimeError("ASGI lifespan task exited during shutdown without a completion message")
                raise RuntimeError("ASGI lifespan task crashed during shutdown") from exc

            msg_type = message["type"]

            if msg_type == "lifespan.shutdown.complete":
                await self._await_task(
                    session.task,
                    timeout=self.shutdown_timeout,
                    timeout_message="ASGI lifespan app did not exit after shutdown.complete",
                )
                return

            await self._cancel_task(session.task)

            if msg_type == "lifespan.shutdown.failed":
                detail = message.get("message", "")
                raise RuntimeError(f"ASGI lifespan shutdown failed: {detail}")

            raise RuntimeError(f"Unexpected ASGI lifespan shutdown message: {msg_type!r}")



from __future__ import annotations

from typing import (
    Callable,
    cast,
)

import typing_extensions

from zapros._handlers._common import (
    ensure_async_handler,
    ensure_sync_handler,
)
from zapros._handlers._sync_base import (
    BaseHandler,
    BaseMiddleware,
)
from zapros.matchers import Matcher

from .._models import (
    Request,
    Response,
)
from ._async_base import (
    AsyncBaseHandler,
    AsyncBaseMiddleware,
)


class Mock:
    def __init__(self) -> None:
        self.matchers: list[Matcher] = []
        self._response: Response | None = None
        self._callback: Callable[[Request], Response] | BaseException | type[BaseException] | None = None
        self._expected_calls: int | None = None
        self._actual_calls: int = 0
        self._name: str | None = None

    @staticmethod
    def given(matcher: Matcher) -> Mock:
        mock = Mock()
        mock.matchers.append(matcher)
        return mock

    def and_(self, matcher: Matcher) -> Mock:
        self.matchers.append(matcher)
        return self

    def mount(self, router: MockRouter) -> Mock:
        router.add(self)
        return self

    def respond(self, response: Response) -> Mock:
        self._response = response
        return self

    def callback(
        self,
        fn: Callable[[Request], Response] | BaseException | type[BaseException],
    ) -> Mock:
        self._callback = fn
        return self

    def expect(self, n: int) -> Mock:
        self._expected_calls = n
        return self

    def once(self) -> Mock:
        return self.expect(1)

    def never(self) -> Mock:
        return self.expect(0)

    def name(self, name: str) -> Mock:
        self._name = name
        return self

    def _is_exhausted(self) -> bool:
        return (
            self._expected_calls is not None and self._actual_calls >= self._expected_calls and self._expected_calls > 0
        )

    def matches(self, request: Request) -> bool:
        if self._is_exhausted():
            return False
        return all(m.match(request) for m in self.matchers)

    def handle(self, request: Request) -> Response:
        self._actual_calls += 1

        if self._callback:
            if isinstance(
                self._callback,
                BaseException,
            ):
                raise self._callback
            if isinstance(self._callback, type) and issubclass(
                self._callback,
                BaseException,
            ):
                raise self._callback()
            return self._callback(request)  # type: ignore

        if self._response is not None:
            return self._response

        return Response(status=200)

    def verify(self) -> None:
        if self._expected_calls is None:
            return

        if self._actual_calls != self._expected_calls:
            name = self._name or "Mock"
            raise AssertionError(f"{name}: expected {self._expected_calls} calls, got {self._actual_calls}")

    def reset(self) -> None:
        self._actual_calls = 0


class MockRouter:
    def __init__(self) -> None:
        self.mocks: list[Mock] = []

    def add(self, mock: Mock) -> MockRouter:
        self.mocks.append(mock)
        return self

    def dispatch(self, request: Request) -> Response | None:
        for mock in self.mocks:
            if mock.matches(request):
                return mock.handle(request)
        return None

    def verify(self) -> None:
        for mock in self.mocks:
            mock.verify()

    def reset(self) -> None:
        for mock in self.mocks:
            mock.reset()


class MockMiddleware(AsyncBaseMiddleware, BaseMiddleware):
    def __init__(
        self,
        router: MockRouter,
        next_handler: AsyncBaseHandler | BaseHandler | None = None,
    ) -> None:
        self.next = cast(BaseHandler, next_handler)
        self.async_next = cast(
            AsyncBaseHandler,
            next_handler,
        )
        self._router = router
        self._fallback = next_handler
        self._passthrough = next_handler is not None

    async def ahandle(self, request: Request) -> Response:

        response = self._router.dispatch(request)
        if response is not None:
            return response

        if not self._passthrough:
            raise ValueError(f"No mock matched request: {request.method} {request.url.pathname}")

        handler = ensure_async_handler(self.async_next)
        return await handler.ahandle(request)

    def handle(self, request: Request) -> Response:

        response = self._router.dispatch(request)
        if response is not None:
            return response

        if not self._passthrough:
            raise ValueError(f"No mock matched request: {request.method} {request.url.pathname}")

        handler = ensure_sync_handler(self.next)
        return handler.handle(request)

    async def aclose(self) -> None:
        self._router.verify()
        self._router.reset()

    def close(self) -> None:
        self._router.verify()
        self._router.reset()


@typing_extensions.deprecated(
    "MockHandler is deprecated, use MockMiddleware instead. "
    "The name 'Handler' was misleading as this is a middleware, not a terminal handler."
)
class MockHandler(MockMiddleware):
    pass



from asyncio import sleep

import pytest

from zapros._async_pool import AsyncConnPool


class MockAsyncConnection:
    def __init__(self, conn_id: int, reusable: bool = True, fail_on_close: bool = False):
        self.conn_id = conn_id
        self._closed = False
        self._reusable = reusable
        self._fail_on_close = fail_on_close
        self.close_count = 0

    async def close(self) -> None:
        self.close_count += 1
        self._closed = True
        if self._fail_on_close:
            raise RuntimeError(f"Connection {self.conn_id} failed to close")
        await sleep(0)

    @property
    def is_closed(self) -> bool:
        return self._closed

    def can_reuse(self) -> bool:
        return self._reusable and not self._closed


@pytest.fixture
def pool():
    return AsyncConnPool(
        max_connections_per_host=10,
        max_idle_per_host=5,
        max_idle_seconds=30.0,
    )


async def get_state(pool: AsyncConnPool, key):
    async with pool._lock:
        return pool._states.get(key)


@pytest.mark.asyncio
async def test_acquire_with_no_idle_returns_none(pool):
    key = ("http", "example.com", 80)
    conn = await pool.acquire(key)
    assert conn is None


@pytest.mark.asyncio
async def test_release_and_reacquire_connection(pool):
    key = ("http", "example.com", 80)
    mock_conn = MockAsyncConnection(conn_id=1)

    await pool.release(key, mock_conn, reuse=True)

    conn = await pool.acquire(key)
    assert conn is mock_conn
    assert conn.close_count == 0


@pytest.mark.asyncio
async def test_release_without_reuse_closes_connection(pool):
    key = ("http", "example.com", 80)
    mock_conn = MockAsyncConnection(conn_id=1)

    await pool.release(key, mock_conn, reuse=False)

    conn = await pool.acquire(key)
    assert conn is None
    assert mock_conn.is_closed


@pytest.mark.asyncio
async def test_expired_connection_not_reused(pool):
    pool._max_age = 0.1
    key = ("http", "example.com", 80)
    mock_conn = MockAsyncConnection(conn_id=1)

    await pool.release(key, mock_conn, reuse=True)
    await sleep(0.15)

    conn = await pool.acquire(key)
    assert conn is None
    assert mock_conn.is_closed


@pytest.mark.asyncio
async def test_non_reusable_connection_discarded(pool):
    key = ("http", "example.com", 80)
    mock_conn = MockAsyncConnection(conn_id=1, reusable=False)

    await pool.release(key, mock_conn, reuse=True)

    conn = await pool.acquire(key)
    assert conn is None
    assert mock_conn.is_closed


@pytest.mark.asyncio
async def test_stale_connections_closed_before_returning_reusable(pool):
    pool._max_age = 0.05
    key = ("http", "example.com", 80)

    reusable_conn = MockAsyncConnection(conn_id=1)
    stale_conn = MockAsyncConnection(conn_id=2, reusable=False)

    await pool.release(key, reusable_conn, reuse=True)
    await pool.release(key, stale_conn, reuse=True)

    acquired = await pool.acquire(key)

    assert acquired is reusable_conn
    assert stale_conn.is_closed
    assert stale_conn.close_count == 1


@pytest.mark.asyncio
async def test_max_idle_connections_enforced(pool):
    key = ("http", "example.com", 80)
    connections = [MockAsyncConnection(conn_id=i) for i in range(10)]

    for conn in connections:
        await pool.release(key, conn, reuse=True)

    remaining_conns = []
    for _ in range(10):
        conn = await pool.acquire(key)
        if conn:
            remaining_conns.append(conn)

    assert len(remaining_conns) == pool._max_idle_per_host

    closed_count = sum(1 for c in connections if c.is_closed)
    assert closed_count == 5


# @pytest.mark.asyncio
# async def test_acquire_semaphore_limits_concurrent_connections():
#     pool = AsyncConnPool(max_connections_per_host=2)
#     key = ("http", "example.com", 80)

#     state = await pool._get_state(key)
#     assert state.semaphore._value == 2
#     await pool._release_state_ref(key, state)

#     conn1 = await pool.acquire(key)
#     assert conn1 is None
#     state = await get_state(pool, key)
#     assert state is not None
#     assert state.semaphore._value == 1

#     conn2 = await pool.acquire(key)
#     assert conn2 is None
#     state = await get_state(pool, key)
#     assert state is not None
#     assert state.semaphore._value == 0

#     acquire_task = asyncio.create_task(pool.acquire(key))
#     await sleep(0.01)
#     assert not acquire_task.done()

#     await pool.release_reservation(key)
#     await sleep(0.01)
#     assert acquire_task.done()

#     acquire_task.cancel()


@pytest.mark.asyncio
async def test_semaphore_per_key():
    pool = AsyncConnPool(max_connections_per_host=2)
    key1 = ("http", "example.com", 80)
    key2 = ("http", "other.com", 80)

    await pool.acquire(key1)
    await pool.acquire(key1)

    conn = await pool.acquire(key2)
    assert conn is None


@pytest.mark.asyncio
async def test_close_all_closes_idle_connections(pool):
    key = ("http", "example.com", 80)
    connections = [MockAsyncConnection(conn_id=i) for i in range(5)]

    for conn in connections:
        await pool.release(key, conn, reuse=True)

    await pool.close_all()

    assert all(conn.is_closed for conn in connections)
    assert pool._closed
    assert len(pool._states) == 0


@pytest.mark.asyncio
async def test_acquire_when_closed_raises_error(pool):
    key = ("http", "example.com", 80)
    await pool.close_all()

    with pytest.raises(RuntimeError, match="Pool is closed"):
        await pool.acquire(key)


@pytest.mark.asyncio
async def test_release_when_closed_discards_connection(pool):
    key = ("http", "example.com", 80)
    mock_conn = MockAsyncConnection(conn_id=1)

    await pool.close_all()
    await pool.release(key, mock_conn, reuse=True)

    assert mock_conn.is_closed


@pytest.mark.asyncio
async def test_close_all_idempotent(pool):
    await pool.close_all()
    await pool.close_all()
    assert pool._closed


@pytest.mark.asyncio
async def test_connection_close_failure_handled_gracefully(pool):
    key = ("http", "example.com", 80)
    failing_conn = MockAsyncConnection(conn_id=1, fail_on_close=True)

    await pool.release(key, failing_conn, reuse=False)
    assert failing_conn.close_count == 1


@pytest.mark.asyncio
async def test_multiple_keys_isolated(pool):
    key1 = ("http", "example.com", 80)
    key2 = ("https", "example.com", 443)
    key3 = ("http", "other.com", 80)

    conn1 = MockAsyncConnection(conn_id=1)
    conn2 = MockAsyncConnection(conn_id=2)
    conn3 = MockAsyncConnection(conn_id=3)

    await pool.release(key1, conn1, reuse=True)
    await pool.release(key2, conn2, reuse=True)
    await pool.release(key3, conn3, reuse=True)

    acquired1 = await pool.acquire(key1)
    acquired2 = await pool.acquire(key2)
    acquired3 = await pool.acquire(key3)

    assert acquired1 is conn1
    assert acquired2 is conn2
    assert acquired3 is conn3


@pytest.mark.asyncio
async def test_lifo_order_for_idle_connections(pool):
    key = ("http", "example.com", 80)
    connections = [MockAsyncConnection(conn_id=i) for i in range(3)]

    for conn in connections:
        await pool.release(key, conn, reuse=True)

    acquired_connections = []
    for _ in range(3):
        conn = await pool.acquire(key)
        if conn:
            acquired_connections.append(conn)

    assert acquired_connections == list(reversed(connections))


# @pytest.mark.asyncio
# async def test_concurrent_acquire_release_stress():
#     pool = AsyncConnPool(max_connections_per_host=20, max_idle_per_host=10)
#     key = ("http", "example.com", 80)

#     async def worker(worker_id: int):
#         for iteration in range(10):
#             conn = await pool.acquire(key)
#             if conn is None:
#                 conn = MockAsyncConnection(conn_id=worker_id * 1000 + iteration)
#             else:
#                 await pool.release_reservation(key)

#             await sleep(0.001)
#             await pool.release(key, conn, reuse=True)

#     tasks = [asyncio.create_task(worker(i)) for i in range(50)]
#     await asyncio.gather(*tasks)

#     await pool.close_all()


@pytest.mark.asyncio
async def test_connection_expiry_during_mixed_operations():
    pool = AsyncConnPool(max_connections_per_host=5, max_idle_seconds=0.1)
    key = ("http", "example.com", 80)

    old_conn = MockAsyncConnection(conn_id=1)
    await pool.release(key, old_conn, reuse=True)

    await sleep(0.15)

    new_conn = MockAsyncConnection(conn_id=2)
    await pool.release(key, new_conn, reuse=True)

    acquired = await pool.acquire(key)
    assert acquired is new_conn

    second_acquired = await pool.acquire(key)
    assert second_acquired is None
    assert old_conn.is_closed


@pytest.mark.asyncio
async def test_partial_reusable_connections():
    pool = AsyncConnPool()
    key = ("http", "example.com", 80)

    connections = [
        MockAsyncConnection(conn_id=1, reusable=False),
        MockAsyncConnection(conn_id=2, reusable=True),
        MockAsyncConnection(conn_id=3, reusable=False),
    ]

    for conn in connections:
        await pool.release(key, conn, reuse=True)

    acquired = await pool.acquire(key)
    assert acquired is connections[1]
    assert connections[2].is_closed
    assert not connections[0].is_closed

    second_acquired = await pool.acquire(key)
    assert second_acquired is None
    assert connections[0].is_closed


@pytest.mark.asyncio
async def test_semaphore_release_on_acquire_failure(pool):
    key = ("http", "example.com", 80)

    state = await pool._get_state(key)
    await pool._release_state_ref(key, state)

    await pool.close_all()

    try:
        await pool.acquire(key)
    except RuntimeError:
        pass

    assert len(pool._states) == 0


# @pytest.mark.asyncio
# async def test_many_keys_concurrent_operations():
#     pool = AsyncConnPool(max_connections_per_host=5)
#     keys = [("http", f"host{i}.com", 80) for i in range(20)]

#     async def work_on_key(key):
#         for _ in range(5):
#             conn = await pool.acquire(key)
#             if conn is None:
#                 conn = MockAsyncConnection(conn_id=hash(key))
#             else:
#                 await pool.release_reservation(key)

#             await sleep(0.001)
#             await pool.release(key, conn, reuse=True)

#     tasks = [asyncio.create_task(work_on_key(key)) for key in keys]
#     await asyncio.gather(*tasks)

#     await pool.close_all()


@pytest.mark.asyncio
async def test_rapid_acquire_release_same_connection():
    pool = AsyncConnPool()
    key = ("http", "example.com", 80)
    conn = MockAsyncConnection(conn_id=1)

    for _ in range(100):
        await pool.release(key, conn, reuse=True)
        acquired = await pool.acquire(key)
        assert acquired is conn


@pytest.mark.asyncio
async def test_empty_idle_queue_after_stale_connections(pool):
    pool._max_age = 0.05
    key = ("http", "example.com", 80)

    connections = [MockAsyncConnection(conn_id=i) for i in range(3)]
    for conn in connections:
        await pool.release(key, conn, reuse=True)

    await sleep(0.1)

    acquired = await pool.acquire(key)
    assert acquired is None
    assert all(c.is_closed for c in connections)
    assert key not in pool._states


@pytest.mark.asyncio
async def test_release_reservation_for_nonexistent_key():
    pool = AsyncConnPool()
    key = ("http", "example.com", 80)

    await pool.release_reservation(key)

    assert key not in pool._states


@pytest.mark.asyncio
async def test_close_all_clears_semaphores(pool):
    keys = [("http", f"host{i}.com", 80) for i in range(5)]

    for key in keys:
        await pool.acquire(key)

    assert len(pool._states) == 5

    await pool.close_all()

    assert len(pool._states) == 0


@pytest.mark.asyncio
async def test_semaphore_entry_removed_after_reservation_released():
    pool = AsyncConnPool()
    key = ("http", "example.com", 80)

    acquired = await pool.acquire(key)

    assert acquired is None
    assert key in pool._states

    await pool.release_reservation(key)

    assert key not in pool._states


# @pytest.mark.asyncio
# async def test_high_load_no_deadlock():
#     pool = AsyncConnPool(max_connections_per_host=10, max_idle_per_host=5)
#     key = ("http", "example.com", 80)
#     results = []

#     async def heavy_worker(worker_id: int):
#         for i in range(20):
#             conn = await pool.acquire(key)
#             if conn is None:
#                 conn = MockAsyncConnection(conn_id=worker_id * 100 + i)
#             else:
#                 await pool.release_reservation(key)

#             await sleep(0.0001)
#             results.append((worker_id, i))
#             await pool.release(key, conn, reuse=(i % 2 == 0))

#     tasks = [asyncio.create_task(heavy_worker(i)) for i in range(30)]

#     await asyncio.wait_for(asyncio.gather(*tasks), timeout=5.0)

#     assert len(results) == 600
#     await pool.close_all()


@pytest.mark.asyncio
async def test_acquire_all_connections_then_release():
    pool = AsyncConnPool(max_connections_per_host=3)
    key = ("http", "example.com", 80)

    state = await pool._get_state(key)
    assert state.semaphore._value == 3
    await pool._release_state_ref(key, state)

    acquired = []
    for _ in range(3):
        conn = await pool.acquire(key)
        acquired.append(conn)

    state = await get_state(pool, key)
    assert state is not None
    assert state.semaphore._value == 0

    for _ in range(3):
        await pool.release_reservation(key)

    assert key not in pool._states



import gzip
import importlib.util
import zlib

import pytest

from zapros._decoders import (
    SUPPORTED_DECODERS,
    ByteChunker,
    DecodingError,
    DeflateDecoder,
    GZipDecoder,
    IdentityDecoder,
    MultiDecoder,
)


def test_identity_decoder():
    decoder = IdentityDecoder()
    data = b"Hello, World!"
    assert decoder.decode(data) == data
    assert decoder.flush() == b""


def test_gzip_decoder():
    original = b"Hello, World! This is a test message for gzip compression."
    compressed = gzip.compress(original)

    decoder = GZipDecoder()
    decoded = decoder.decode(compressed)
    decoded += decoder.flush()

    assert decoded == original


def test_gzip_decoder_streaming():
    original = b"Hello, World! This is a test message for gzip compression."
    compressed = gzip.compress(original)

    decoder = GZipDecoder()
    chunk_size = 10
    result = b""

    for i in range(0, len(compressed), chunk_size):
        chunk = compressed[i : i + chunk_size]
        result += decoder.decode(chunk)

    result += decoder.flush()
    assert result == original


def test_gzip_decoder_invalid_data():
    decoder = GZipDecoder()
    with pytest.raises(
        DecodingError,
        match="Failed to decode gzip data",
    ):
        decoder.decode(b"invalid gzip data")


def test_deflate_decoder():
    original = b"Hello, World! This is a test message for deflate compression."
    compressed = zlib.compress(original)

    decoder = DeflateDecoder()
    decoded = decoder.decode(compressed)
    decoded += decoder.flush()

    assert decoded == original


def test_deflate_decoder_raw():
    original = b"Hello, World! This is a test message for raw deflate compression."
    compressor = zlib.compressobj(wbits=-zlib.MAX_WBITS)
    compressed = compressor.compress(original) + compressor.flush()

    decoder = DeflateDecoder()
    decoded = decoder.decode(compressed)
    decoded += decoder.flush()

    assert decoded == original


def test_deflate_decoder_streaming():
    original = b"Hello, World! This is a test message for deflate compression."
    compressed = zlib.compress(original)

    decoder = DeflateDecoder()
    chunk_size = 10
    result = b""

    for i in range(0, len(compressed), chunk_size):
        chunk = compressed[i : i + chunk_size]
        result += decoder.decode(chunk)

    result += decoder.flush()
    assert result == original


def test_deflate_decoder_invalid_data():
    decoder = DeflateDecoder()
    decoder._first_attempt = False
    with pytest.raises(
        DecodingError,
        match="Failed to decode deflate data",
    ):
        decoder.decode(b"invalid deflate data that is long enough")


def test_brotli_decoder():
    try:
        import brotli
    except ImportError:
        try:
            import brotlicffi as brotli  # type: ignore[import-not-found, no-redef]
        except ImportError:
            pytest.skip("brotli not installed")

    from zapros._decoders import (
        BrotliDecoder,
    )

    original = b"Hello, World! This is a test message for brotli compression."
    compressed = brotli.compress(original)  # type: ignore[attr-defined]

    decoder = BrotliDecoder()
    decoded = decoder.decode(compressed)
    decoded += decoder.flush()

    assert decoded == original


def test_brotli_decoder_streaming():
    try:
        import brotli
    except ImportError:
        try:
            import brotlicffi as brotli  # type: ignore[import-not-found, no-redef]
        except ImportError:
            pytest.skip("brotli not installed")

    from zapros._decoders import (
        BrotliDecoder,
    )

    original = b"Hello, World! This is a test message for brotli compression."
    compressed = brotli.compress(original)

    decoder = BrotliDecoder()
    chunk_size = 10
    result = b""

    for i in range(0, len(compressed), chunk_size):
        chunk = compressed[i : i + chunk_size]
        result += decoder.decode(chunk)

    result += decoder.flush()
    assert result == original


def test_zstandard_decoder():
    try:
        import zstandard
    except ImportError:
        pytest.skip("zstandard not installed")

    from zapros._decoders import (
        ZStandardDecoder,
    )

    original = b"Hello, World! This is a test message for zstd compression."
    compressor = zstandard.ZstdCompressor()
    compressed = compressor.compress(original)

    decoder = ZStandardDecoder()
    decoded = decoder.decode(compressed)
    decoded += decoder.flush()

    assert decoded == original


def test_zstandard_decoder_streaming():
    try:
        import zstandard
    except ImportError:
        pytest.skip("zstandard not installed")

    from zapros._decoders import (
        ZStandardDecoder,
    )

    original = b"Hello, World! This is a test message for zstd compression."
    compressor = zstandard.ZstdCompressor()
    compressed = compressor.compress(original)

    decoder = ZStandardDecoder()
    chunk_size = 10
    result = b""

    for i in range(0, len(compressed), chunk_size):
        chunk = compressed[i : i + chunk_size]
        result += decoder.decode(chunk)

    result += decoder.flush()
    assert result == original


def test_multi_decoder():
    original = b"Hello, World! This is a test message for multi-layer compression."

    compressed_gzip = gzip.compress(original)
    compressed_deflate = zlib.compress(compressed_gzip)

    decoder = MultiDecoder(
        [
            GZipDecoder(),
            DeflateDecoder(),
        ]
    )
    decoded = decoder.decode(compressed_deflate)
    decoded += decoder.flush()

    assert decoded == original


def test_multi_decoder_streaming():
    original = b"Hello, World! This is a test message for multi-layer compression."

    compressed_gzip = gzip.compress(original)
    compressed_deflate = zlib.compress(compressed_gzip)

    decoder = MultiDecoder(
        [
            GZipDecoder(),
            DeflateDecoder(),
        ]
    )
    chunk_size = 20
    result = b""

    for i in range(
        0,
        len(compressed_deflate),
        chunk_size,
    ):
        chunk = compressed_deflate[i : i + chunk_size]
        result += decoder.decode(chunk)

    result += decoder.flush()
    assert result == original


def test_byte_chunker():
    chunker = ByteChunker(chunk_size=10)

    data = b"Hello, World! This is a test."
    chunks = list(chunker.feed(data))

    assert len(chunks) == 2
    assert chunks[0] == b"Hello, Wor"
    assert chunks[1] == b"ld! This i"

    remaining = chunker.flush()
    assert remaining == b"s a test."


def test_byte_chunker_exact_size():
    chunker = ByteChunker(chunk_size=10)

    data = b"1234567890"
    chunks = list(chunker.feed(data))

    assert len(chunks) == 1
    assert chunks[0] == b"1234567890"

    remaining = chunker.flush()
    assert remaining == b""


def test_byte_chunker_multiple_feeds():
    chunker = ByteChunker(chunk_size=10)

    all_chunks = []
    all_chunks.extend(chunker.feed(b"Hello, "))
    all_chunks.extend(chunker.feed(b"World! "))
    all_chunks.extend(chunker.feed(b"This is a test."))

    assert len(all_chunks) == 2
    assert all_chunks[0] == b"Hello, Wor"
    assert all_chunks[1] == b"ld! This i"

    remaining = chunker.flush()
    assert remaining == b"s a test."


def test_byte_chunker_empty():
    chunker = ByteChunker(chunk_size=10)
    chunks = list(chunker.feed(b""))
    assert chunks == []

    remaining = chunker.flush()
    assert remaining == b""


def test_supported_decoders_registry():
    assert "identity" in SUPPORTED_DECODERS
    assert "gzip" in SUPPORTED_DECODERS
    assert "deflate" in SUPPORTED_DECODERS

    assert SUPPORTED_DECODERS["identity"] == IdentityDecoder
    assert SUPPORTED_DECODERS["gzip"] == GZipDecoder
    assert SUPPORTED_DECODERS["deflate"] == DeflateDecoder


def test_brotli_in_registry():
    has_brotli = importlib.util.find_spec("brotli") is not None
    has_brotlicffi = importlib.util.find_spec("brotlicffi") is not None
    if has_brotli or has_brotlicffi:
        assert "br" in SUPPORTED_DECODERS
    else:
        assert "br" not in SUPPORTED_DECODERS


def test_zstandard_in_registry():
    has_zstandard = importlib.util.find_spec("zstandard") is not None
    if has_zstandard:
        assert "zstd" in SUPPORTED_DECODERS
    else:
        assert "zstd" not in SUPPORTED_DECODERS



from __future__ import annotations

from abc import abstractmethod
from typing import (
    Callable,
    Protocol,
    runtime_checkable,
)

from .._models import Request, Response


@runtime_checkable
class AsyncBaseHandler(Protocol):
    @abstractmethod
    async def ahandle(self, request: Request) -> Response:
        raise NotImplementedError()

    async def aclose(self) -> None:
        pass

    def wrap_with_async_middleware(
        self,
        factory: Callable[
            [AsyncBaseHandler],
            AsyncBaseMiddleware,
        ],
    ) -> AsyncBaseMiddleware:
        return factory(self)


@runtime_checkable
class AsyncBaseMiddleware(AsyncBaseHandler, Protocol):
    async_next: AsyncBaseHandler



from __future__ import annotations

from collections.abc import (
    Iterable,
    Iterator,
    Mapping,
)
from typing import (
    Generic,
    MutableMapping,
    Protocol,
    TypeVar,
    cast,
)

_VT = TypeVar("_VT")
_T = TypeVar("_T")
_KT = TypeVar("_KT")
_VT_co = TypeVar("_VT_co", covariant=True)


class SupportsKeysAndGetItem(Protocol[_KT, _VT_co]):
    def keys(self) -> Iterable[_KT]: ...
    def __getitem__(self, __key: _KT, /) -> _VT_co: ...


class CIMultiDict(MutableMapping[str, _VT], Generic[_VT]):
    __slots__ = ("_items", "_index")

    def __init__(
        self,
        arg: Mapping[str, _VT] | Iterable[tuple[str, _VT]] | None = None,
    ) -> None:
        self._items: list[tuple[str, _VT]] = []
        self._index: dict[str, list[int]] = {}

        if arg is not None:
            self.extend(arg)

    @staticmethod
    def _norm_key(key: str) -> str:
        return key.casefold()

    def _reindex(self) -> None:
        index: dict[str, list[int]] = {}
        for i, (key, _) in enumerate(self._items):
            norm: str = self._norm_key(key)
            index.setdefault(norm, []).append(i)
        self._index = index

    def __getitem__(self, key: str) -> _VT:
        norm: str = self._norm_key(key)
        try:
            first_pos: int = self._index[norm][0]
        except KeyError:
            raise KeyError(key) from None
        return self._items[first_pos][1]

    def __setitem__(self, key: str, value: _VT) -> None:
        norm: str = self._norm_key(key)
        self._items = [(k, v) for k, v in self._items if self._norm_key(k) != norm]
        self._items.append((key, value))
        self._reindex()

    def __delitem__(self, key: str) -> None:
        norm: str = self._norm_key(key)
        items: list[tuple[str, _VT]] = [(k, v) for k, v in self._items if self._norm_key(k) != norm]
        if len(items) == len(self._items):
            raise KeyError(key)
        self._items = items
        self._reindex()

    def __contains__(self, key: object) -> bool:
        return isinstance(key, str) and self._norm_key(key) in self._index

    def __iter__(self) -> Iterator[str]:
        seen: set[str] = set()
        for key, _ in self._items:
            norm: str = self._norm_key(key)
            if norm not in seen:
                seen.add(norm)
                yield key

    def __len__(self) -> int:
        return len(self._index)

    def __repr__(self) -> str:
        return f"<CIMultiDict({self._items!r})>"

    def getall(self, key: str) -> list[_VT]:
        norm: str = self._norm_key(key)
        positions: list[int] | None = self._index.get(norm)
        if not positions:
            raise KeyError(key)
        return [self._items[pos][1] for pos in positions]

    def allitems(self) -> Iterator[tuple[str, _VT]]:
        yield from self._items

    def add(self, key: str, value: _VT) -> None:
        pos: int = len(self._items)
        self._items.append((key, value))
        norm: str = self._norm_key(key)
        self._index.setdefault(norm, []).append(pos)

    def extend(
        self,
        arg: SupportsKeysAndGetItem[str, _VT] | Iterable[tuple[str, _VT]],
    ) -> None:
        items: Iterable[tuple[str, _VT]]
        if isinstance(arg, Mapping):
            items = cast(Iterable[tuple[str, _VT]], arg.items())
        elif hasattr(arg, "keys"):
            arg_keyed: SupportsKeysAndGetItem[str, _VT] = cast(
                SupportsKeysAndGetItem[str, _VT],
                arg,
            )
            items = [(k, arg_keyed[k]) for k in arg_keyed.keys()]
        else:
            items = cast(Iterable[tuple[str, _VT]], arg)

        for key, value in items:
            self.add(key, value)

    def popitem(self) -> tuple[str, _VT]:
        if not self._items:
            raise KeyError("popitem(): dictionary is empty")
        key: str
        value: _VT
        key, value = self._items[-1]
        del self[key]
        return key, value

    def copy(self) -> CIMultiDict[_VT]:
        new: CIMultiDict[_VT] = type(self)()
        new._items = self._items.copy()
        new._index = {k: v.copy() for k, v in self._index.items()}
        return new
